# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import pytest

from raillabel.format import Size2d

# == Fixtures =========================


@pytest.fixture
def size2d_json() -> dict:
    return [25, 1.344]


@pytest.fixture
def size2d() -> dict:
    return Size2d(25, 1.344)


# == Tests ============================


def test_from_json(size2d, size2d_json):
    actual = Size2d.from_json(size2d_json)
    assert actual == size2d


def test_to_json(size2d, size2d_json):
    actual = size2d.to_json()
    assert actual == tuple(size2d_json)


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from decimal import Decimal
from uuid import UUID

from ._json_format_base import _JSONFormatBase
from .num import JSONNum
from .object_data import JSONObjectData
from .stream_sync import JSONStreamSync


class JSONFrame(_JSONFormatBase):
    """A frame is a container of dynamic, timewise, information."""

    frame_properties: JSONFrameProperties | None = None
    "Container of frame information other than annotations."

    objects: dict[UUID, JSONObjectData] | None = None
    """This is a JSON object that contains dynamic information on RailLabel objects. Object keys are
    strings containing 32 bytes UUIDs. Object values contain an 'object_data' JSON object."""


class JSONFrameProperties(_JSONFormatBase):
    """Container of frame information other than annotations."""

    timestamp: Decimal | str | None = None
    """The timestamp indicates a time instant as a UTC string or numerical value to describe this
    frame."""

    streams: dict[str, JSONStreamSync] | None = None
    """Stream timestamps for synchronization."""

    frame_data: JSONFrameData | None = None
    "Additional data to describe attributes of the frame (like GPS position)."


class JSONFrameData(_JSONFormatBase):
    """Additional data to describe attributes of the frame (like GPS position)."""

    num: list[JSONNum] | None = None
    "List of 'num' that describe this frame."



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from dataclasses import dataclass


@dataclass
class Point2d:
    """A 2d point in an image."""

    x: float
    "The x-coordinate of the point in the image in pixels from the left."

    y: float
    "The y-coordinate of the point in the image in pixels from the top."

    @classmethod
    def from_json(cls, json: tuple[float, float]) -> Point2d:
        """Construct an instant of this class from RailLabel JSON data."""
        return Point2d(x=json[0], y=json[1])

    def to_json(self) -> tuple[float, float]:
        """Export this object into the RailLabel JSON format."""
        return (self.x, self.y)



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import pytest

from raillabel.format import Radar, IntrinsicsRadar
from raillabel.json_format import (
    JSONCoordinateSystem,
    JSONStreamRadar,
    JSONStreamRadarProperties,
    JSONIntrinsicsRadar,
)

# == Fixtures =========================


@pytest.fixture
def radar_json(
    intrinsics_radar_json, transform_json
) -> tuple[JSONStreamRadar, JSONCoordinateSystem]:
    return (
        JSONStreamRadar(
            type="radar",
            stream_properties=JSONStreamRadarProperties(intrinsics_radar=intrinsics_radar_json),
            uri="/path/to/sensor/data",
            description="A very nice radar",
        ),
        JSONCoordinateSystem(
            parent="base", type="sensor", pose_wrt_parent=transform_json, children=None
        ),
    )


@pytest.fixture
def radar(intrinsics_radar, transform) -> Radar:
    return Radar(
        intrinsics=intrinsics_radar,
        extrinsics=transform,
        uri="/path/to/sensor/data",
        description="A very nice radar",
    )


@pytest.fixture
def radar_empty() -> Radar:
    return Radar(
        intrinsics=IntrinsicsRadar(
            resolution_px_per_m=0,
            width_px=0,
            height_px=0,
        )
    )


# == Tests ============================


def test_from_json(radar, radar_json):
    actual = Radar.from_json(radar_json[0], radar_json[1])
    assert actual == radar


def test_to_json(radar, radar_json):
    actual = radar.to_json()
    assert actual == radar_json


if __name__ == "__main__":
    pytest.main([__file__, "-vv"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

import os

import pytest

from raillabel.json_format import JSONScene


def test_parse__openlabel_short(json_data):
    JSONScene(**json_data["openlabel_v1_short"])


# Executes the test if the file is called
if __name__ == "__main__":
    os.system("clear")
    pytest.main([__file__, "--disable-pytest-warnings", "--cache-clear"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from dataclasses import dataclass

from raillabel.json_format import (
    JSONCoordinateSystem,
    JSONStreamCamera,
    JSONStreamCameraProperties,
    JSONTransformData,
)

from .intrinsics_pinhole import IntrinsicsPinhole
from .transform import Transform


@dataclass
class Camera:
    """A camera sensor."""

    intrinsics: IntrinsicsPinhole
    "The intrinsic calibration of the sensor."

    extrinsics: Transform | None = None
    "External calibration of the sensor defined by the 3D transform to the coordinate system origin."

    uri: str | None = None
    "Name of the subdirectory containing the sensor files."

    description: str | None = None
    "Additional information about the sensor."

    TYPE: str = "camera"

    @classmethod
    def from_json(
        cls, json_stream: JSONStreamCamera, json_coordinate_system: JSONCoordinateSystem
    ) -> Camera:
        """Construct an instant of this class from RailLabel JSON data."""
        return Camera(
            intrinsics=IntrinsicsPinhole.from_json(json_stream.stream_properties.intrinsics_pinhole),
            extrinsics=_extrinsics_from_json(json_coordinate_system.pose_wrt_parent),
            uri=json_stream.uri,
            description=json_stream.description,
        )

    def to_json(self) -> tuple[JSONStreamCamera, JSONCoordinateSystem]:
        """Export this object into the RailLabel JSON format."""
        return (
            JSONStreamCamera(
                type="camera",
                stream_properties=JSONStreamCameraProperties(
                    intrinsics_pinhole=self.intrinsics.to_json()
                ),
                uri=self.uri,
                description=self.description,
            ),
            JSONCoordinateSystem(
                parent="base",
                type="sensor",
                pose_wrt_parent=self.extrinsics.to_json() if self.extrinsics is not None else None,
            ),
        )


def _extrinsics_from_json(json_transform: JSONTransformData | None) -> Transform | None:
    if json_transform is None:
        return None
    return Transform.from_json(json_transform)



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from uuid import UUID

from ._json_format_base import _JSONFormatBase
from .attributes import JSONAttributes


class JSONBbox(_JSONFormatBase):
    """A 2D bounding box is defined as a 4-dimensional vector [x, y, w, h].

    [x, y] is the center of the bounding box and [w, h] represent the width (horizontal,
    x-coordinate dimension) and height (vertical, y-coordinate dimension), respectively.
    """

    name: str
    """This is a string encoding the name of this object data. It is used as index inside the
    corresponding object data pointers."""

    val: tuple[float, float, float, float]
    "The array of 4 values that define the [x, y, w, h] values of the bbox."

    coordinate_system: str
    "Name of the coordinate system in respect of which this object data is expressed."

    uid: UUID | None = None
    "This is a string encoding the Universal Unique identifyer of the annotation."

    attributes: JSONAttributes | None = None



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from typing import Literal

from ._json_format_base import _JSONFormatBase


class JSONStreamOther(_JSONFormatBase):
    """A stream describes the source of a data sequence, usually a sensor.

    This specific object describes a sensor without intrinsic calibration.
    """

    type: Literal["lidar", "gps_imu", "other"]
    "A string encoding the type of the stream."

    uri: str | None = None
    "A string encoding the subdirectory containing the sensor files."

    description: str | None = None
    "Description of the stream."



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from copy import deepcopy
from dataclasses import dataclass, field
from uuid import UUID

from raillabel.filter._filter_abc import _AnnotationLevelFilter, _FilterAbc, _FrameLevelFilter
from raillabel.json_format import (
    JSONCoordinateSystem,
    JSONFrame,
    JSONObject,
    JSONScene,
    JSONSceneContent,
    JSONStreamCamera,
    JSONStreamOther,
    JSONStreamRadar,
)

from .bbox import Bbox
from .camera import Camera
from .cuboid import Cuboid
from .frame import Frame
from .frame_interval import FrameInterval
from .gps_imu import GpsImu
from .lidar import Lidar
from .metadata import Metadata
from .object import Object
from .other_sensor import OtherSensor
from .poly2d import Poly2d
from .poly3d import Poly3d
from .radar import Radar
from .seg3d import Seg3d


@dataclass
class Scene:
    """The root RailLabel class, which contains all data."""

    metadata: Metadata
    "Container of information about the annotation file itself."

    sensors: dict[str, Camera | Lidar | Radar | GpsImu | OtherSensor] = field(default_factory=dict)
    "The sensors used in this scene. Keys are sensor names."

    objects: dict[UUID, Object] = field(default_factory=dict)
    "Unique objects (like a specific person) in this scene. Keys are object uuids"

    frames: dict[int, Frame] = field(default_factory=dict)
    "A container of dynamic, timewise, information. Keys are the frame integer number."

    @classmethod
    def from_json(cls, json: JSONScene) -> Scene:
        """Construct a scene from a json object."""
        return Scene(
            metadata=Metadata.from_json(json.openlabel.metadata),
            sensors=_sensors_from_json(json.openlabel.streams, json.openlabel.coordinate_systems),
            objects=_objects_from_json(json.openlabel.objects),
            frames=_frames_from_json(json.openlabel.frames),
        )

    def to_json(self) -> JSONScene:
        """Export this scene into the RailLabel JSON format."""
        return JSONScene(
            openlabel=JSONSceneContent(
                metadata=self.metadata.to_json(),
                streams={
                    sensor_id: sensor.to_json()[0] for sensor_id, sensor in self.sensors.items()
                },
                coordinate_systems=_coordinate_systems_to_json(self.sensors),
                objects={
                    obj_id: obj.to_json(obj_id, self.frames) for obj_id, obj in self.objects.items()
                },
                frames={
                    frame_id: frame.to_json(self.objects) for frame_id, frame in self.frames.items()
                },
                frame_intervals=[
                    fi.to_json() for fi in FrameInterval.from_frame_ids(list(self.frames.keys()))
                ],
            )
        )

    def filter(self, filters: list[_FilterAbc]) -> Scene:
        """Return a scene with annotations, sensors, objects and frames excluded."""
        frame_filters, annotation_filters = _separate_filters(filters)

        filtered_scene = Scene(metadata=deepcopy(self.metadata))
        filtered_scene.frames = _filter_frames(self, frame_filters, annotation_filters)
        filtered_scene.sensors = _get_used_sensors(self, filtered_scene)
        filtered_scene.objects = _get_used_objects(self, filtered_scene)

        return filtered_scene


def _sensors_from_json(
    json_streams: dict[str, JSONStreamCamera | JSONStreamOther | JSONStreamRadar] | None,
    json_coordinate_systems: dict[str, JSONCoordinateSystem] | None,
) -> dict[str, Camera | Lidar | Radar | GpsImu | OtherSensor]:
    sensors: dict[str, Camera | Lidar | Radar | GpsImu | OtherSensor] = {}

    if json_streams is None or json_coordinate_systems is None:
        return sensors

    for sensor_id, json_stream in json_streams.items():
        json_coordinate_system = json_coordinate_systems[sensor_id]

        if isinstance(json_stream, JSONStreamCamera):
            sensors[sensor_id] = Camera.from_json(json_stream, json_coordinate_system)

        if isinstance(json_stream, JSONStreamRadar):
            sensors[sensor_id] = Radar.from_json(json_stream, json_coordinate_system)

        if isinstance(json_stream, JSONStreamOther) and json_stream.type == "lidar":
            sensors[sensor_id] = Lidar.from_json(json_stream, json_coordinate_system)

        if isinstance(json_stream, JSONStreamOther) and json_stream.type == "gps_imu":
            sensors[sensor_id] = GpsImu.from_json(json_stream, json_coordinate_system)

        if isinstance(json_stream, JSONStreamOther) and json_stream.type == "other":
            sensors[sensor_id] = OtherSensor.from_json(json_stream, json_coordinate_system)

    return sensors


def _objects_from_json(json_objects: dict[UUID, JSONObject] | None) -> dict[UUID, Object]:
    if json_objects is None:
        return {}

    return {obj_id: Object.from_json(json_obj) for obj_id, json_obj in json_objects.items()}


def _frames_from_json(json_frames: dict[int, JSONFrame] | None) -> dict[int, Frame]:
    if json_frames is None:
        return {}

    return {frame_id: Frame.from_json(json_frame) for frame_id, json_frame in json_frames.items()}


def _coordinate_systems_to_json(
    sensors: dict[str, Camera | Lidar | Radar | GpsImu | OtherSensor],
) -> dict[str, JSONCoordinateSystem]:
    json_coordinate_systems = {
        sensor_id: sensor.to_json()[1] for sensor_id, sensor in sensors.items()
    }
    json_coordinate_systems["base"] = JSONCoordinateSystem(
        parent="",
        type="local",
        pose_wrt_parent=None,
        children=list(json_coordinate_systems.keys()),
    )
    return json_coordinate_systems


def _separate_filters(
    all_filters: list[_FilterAbc],
) -> tuple[list[_FrameLevelFilter], list[_AnnotationLevelFilter]]:
    frame_filters = []
    annotation_filters = []
    for filter_ in all_filters:
        if isinstance(filter_, _FrameLevelFilter):
            frame_filters.append(filter_)

        if isinstance(filter_, _AnnotationLevelFilter):
            annotation_filters.append(filter_)

    return frame_filters, annotation_filters


def _filter_frames(
    scene: Scene,
    frame_filters: list[_FrameLevelFilter],
    annotation_filters: list[_AnnotationLevelFilter],
) -> dict[int, Frame]:
    filtered_frames = {}

    for frame_id, frame in scene.frames.items():
        if _frame_passes_all_filters(frame_id, frame, frame_filters):
            filtered_frames[frame_id] = Frame(
                timestamp=deepcopy(frame.timestamp),
                sensors=deepcopy(frame.sensors),
                frame_data=deepcopy(frame.frame_data),
                annotations=_filter_annotations(frame, annotation_filters, scene),
            )

    return filtered_frames


def _filter_annotations(
    frame: Frame, annotation_filters: list[_AnnotationLevelFilter], scene: Scene
) -> dict[UUID, Bbox | Cuboid | Poly2d | Poly3d | Seg3d]:
    annotations = {}

    for annotation_id, annotation in frame.annotations.items():
        if _annotation_passes_all_filters(annotation_id, annotation, annotation_filters, scene):
            annotations[annotation_id] = deepcopy(annotation)

    return annotations


def _frame_passes_all_filters(
    frame_id: int, frame: Frame, frame_filters: list[_FrameLevelFilter]
) -> bool:
    return all(filter_.passes_filter(frame_id, frame) for filter_ in frame_filters)


def _annotation_passes_all_filters(
    annotation_id: UUID,
    annotation: Bbox | Cuboid | Poly2d | Poly3d | Seg3d,
    annotation_filters: list[_AnnotationLevelFilter],
    scene: Scene,
) -> bool:
    return all(
        filter_.passes_filter(annotation_id, annotation, scene) for filter_ in annotation_filters
    )


def _get_used_sensors(
    scene: Scene, filtered_scene: Scene
) -> dict[str, Camera | Lidar | Radar | GpsImu | OtherSensor]:
    used_sensors = {}
    for frame in filtered_scene.frames.values():
        for annotation in frame.annotations.values():
            if annotation.sensor_id not in used_sensors:
                used_sensors[annotation.sensor_id] = deepcopy(scene.sensors[annotation.sensor_id])

    return used_sensors


def _get_used_objects(scene: Scene, filtered_scene: Scene) -> dict[UUID, Object]:
    used_objects = {}
    for frame in filtered_scene.frames.values():
        for annotation in frame.annotations.values():
            if annotation.object_id not in used_objects:
                used_objects[annotation.object_id] = deepcopy(scene.objects[annotation.object_id])

    return used_objects



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import json
from pathlib import Path

from raillabel.format import Scene
from raillabel.json_format import JSONScene


def load(path: Path | str) -> Scene:
    """Load an annotation file as a scene."""
    with Path(path).open() as annotation_file:
        json_data = json.load(annotation_file)
    return Scene.from_json(JSONScene(**json_data))



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from dataclasses import dataclass
from uuid import UUID

from raillabel.format import Bbox, Cuboid, Poly2d, Poly3d, Scene, Seg3d

from ._filter_abc import _AnnotationLevelFilter


@dataclass
class IncludeAttributesFilter(_AnnotationLevelFilter):
    """Filter out all annotations in the scene, that do NOT have matching attributes.

    If an attribute has None as the value, all annotations are excluded, that do not have this
    attribute. If the value is anything other than None, all annotations are excluded that do not
    have the attribute or where the value does not match.
    """

    attributes: dict[str, bool | float | str | list | None]

    def passes_filter(
        self, _: UUID, annotation: Bbox | Cuboid | Poly2d | Poly3d | Seg3d, __: Scene
    ) -> bool:
        """Assess if an annotation passes this filter."""
        for attribute_name, attribute_value in annotation.attributes.items():
            if attribute_name not in self.attributes:
                return False

            if (
                self.attributes[attribute_name] is not None
                and attribute_value != self.attributes[attribute_name]
            ):
                return False

        return True



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0
"""Devkit for working with recorded and annotated train ride data from DB."""

from importlib import metadata

from . import filter, format
from .format import Scene
from .load.load import load
from .save.save import save

__all__ = [
    "filter",
    "format",
    "Scene",
    "load",
    "save",
]

try:
    __version__ = metadata.version("raillabel")
except metadata.PackageNotFoundError:
    __version__ = "0.0.0+unknown"
del metadata



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from uuid import UUID

import pytest

from raillabel.format import Poly3d
from raillabel.json_format import JSONPoly3d

# == Fixtures =========================


@pytest.fixture
def poly3d_json(
    point3d_json,
    another_point3d_json,
    attributes_multiple_types_json,
) -> JSONPoly3d:
    return JSONPoly3d(
        uid="0da87210-46F1-40e5-b661-20ea1c392f50",
        name="lidar__poly3d__track",
        closed=True,
        val=point3d_json + another_point3d_json,
        coordinate_system="lidar",
        attributes=attributes_multiple_types_json,
    )


@pytest.fixture
def poly3d_id() -> UUID:
    return UUID("0da87210-46F1-40e5-b661-20ea1c392f50")


@pytest.fixture
def poly3d(
    point3d,
    another_point3d,
    attributes_multiple_types,
    object_track_id,
) -> Poly3d:
    return Poly3d(
        points=[point3d, another_point3d],
        closed=True,
        sensor_id="lidar",
        attributes=attributes_multiple_types,
        object_id=object_track_id,
    )


# == Tests ============================


def test_from_json(poly3d, poly3d_json, object_track_id):
    actual = Poly3d.from_json(poly3d_json, object_track_id)
    assert actual == poly3d


def test_name(poly3d):
    actual = poly3d.name("track")
    assert actual == "lidar__poly3d__track"


def test_to_json(poly3d, poly3d_json, poly3d_id):
    actual = poly3d.to_json(poly3d_id, object_type="track")
    assert actual == poly3d_json


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations


def _empty_list_to_none(collection: list | None) -> list | None:
    if collection is None:
        return None
    if len(collection) == 0:
        return None
    return collection


def _flatten_list(list_of_tuples: list[tuple]) -> list:
    return [item for tup in list_of_tuples for item in tup]



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import glob
import json
import sys
import typing as t
from pathlib import Path

import json5
import pytest

sys.path.insert(1, str(Path(__file__).parent.parent))
import raillabel


from format.test_attributes import attributes_multiple_types, attributes_multiple_types_json
from format.test_bbox import bbox, bbox_json, bbox_id
from format.test_camera import camera, camera_json, camera_empty
from format.test_cuboid import cuboid, cuboid_json, cuboid_id
from format.test_frame import frame, frame_json
from format.test_frame_interval import frame_interval, frame_interval_json
from format.test_intrinsics_pinhole import intrinsics_pinhole, intrinsics_pinhole_json
from format.test_intrinsics_radar import intrinsics_radar, intrinsics_radar_json
from format.test_lidar import lidar, lidar_json
from format.test_metadata import metadata, metadata_json
from format.test_num import num, num_json, num_id
from format.test_object import (
    objects,
    object_person,
    object_person_json,
    object_person_id,
    object_track,
    object_track_json,
    object_track_id,
)
from format.test_point2d import point2d, point2d_json, another_point2d, another_point2d_json
from format.test_point3d import point3d, point3d_json, another_point3d, another_point3d_json
from format.test_poly2d import poly2d, poly2d_json, poly2d_id
from format.test_poly3d import poly3d, poly3d_json, poly3d_id
from format.test_quaternion import quaternion, quaternion_json
from format.test_radar import radar, radar_json, radar_empty
from format.test_size2d import size2d, size2d_json
from format.test_size3d import size3d, size3d_json
from format.test_seg3d import seg3d, seg3d_json, seg3d_id
from format.test_sensor_reference import (
    another_sensor_reference,
    another_sensor_reference_json,
    sensor_reference,
    sensor_reference_json,
)
from format.test_transform import transform, transform_json


json_data_directories = [
    Path(__file__).parent / "__test_assets__",
]


@pytest.fixture(scope="session", autouse=True)
def compile_uncommented_test_file():
    """Compiles the main test file from json5 to json."""

    json5_file_path = json_data_directories[0] / "openlabel_v1_short.json5"

    with json5_file_path.open() as f:
        data = json5.load(f)

    with open(str(json5_file_path)[:-1], "w") as f:
        json.dump(data, f, indent=4)


@pytest.fixture
def json_paths(request) -> dict[str, Path]:
    json_paths = _fetch_json_paths_from_cache(request)

    if json_paths is None:
        json_paths = {_get_file_identifier(p): p for p in _collect_json_paths()}

    return json_paths


def _fetch_json_paths_from_cache(request) -> dict[str, Path] | None:
    return request.config.cache.get("json_paths", None)


def _collect_json_paths() -> list[Path]:
    json_paths = []

    for dir in json_data_directories:
        json_paths.extend([Path(p) for p in glob.glob(str(dir) + "/**/**.json", recursive=True)])

    return json_paths


def _get_file_identifier(path: Path) -> str:
    """Return relative path from test asset dir as string."""

    if "__test_assets__" not in path.parts:
        return path.stem

    test_assets_dir_index = path.parts.index("__test_assets__")

    relative_path = ""
    for part in path.parts[test_assets_dir_index + 1 : -1]:
        relative_path += part + "/"

    relative_path += path.stem

    return relative_path


@pytest.fixture
def json_data(request) -> dict[str, dict]:
    json_data = _fetch_json_data_from_cache(request)

    if json_data is None:
        json_data = {_get_file_identifier(p): _load_json_data(p) for p in _collect_json_paths()}

    return json_data


def _fetch_json_data_from_cache(request) -> dict[str, Path] | None:
    return request.config.cache.get("json_data", None)


def _load_json_data(path: Path) -> dict:
    with path.open() as f:
        json_data = json.load(f)
    return json_data



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0
"""Package for classes representing the direct RailLabel JSON objects."""

from .attributes import JSONAttributes
from .bbox import JSONBbox
from .boolean_attribute import JSONBooleanAttribute
from .coordinate_system import JSONCoordinateSystem
from .cuboid import JSONCuboid
from .element_data_pointer import JSONElementDataPointer
from .frame import JSONFrame, JSONFrameData, JSONFrameProperties
from .frame_interval import JSONFrameInterval
from .metadata import JSONMetadata
from .num import JSONNum
from .num_attribute import JSONNumAttribute
from .object import JSONObject
from .object_data import JSONAnnotations, JSONObjectData
from .poly2d import JSONPoly2d
from .poly3d import JSONPoly3d
from .scene import JSONScene, JSONSceneContent
from .stream_camera import JSONIntrinsicsPinhole, JSONStreamCamera, JSONStreamCameraProperties
from .stream_other import JSONStreamOther
from .stream_radar import JSONIntrinsicsRadar, JSONStreamRadar, JSONStreamRadarProperties
from .stream_sync import JSONStreamSync, JSONStreamSyncProperties, JSONStreamSyncTimestamp
from .text_attribute import JSONTextAttribute
from .transform_data import JSONTransformData
from .vec import JSONVec
from .vec_attribute import JSONVecAttribute

__all__ = [
    "JSONAnnotations",
    "JSONAttributes",
    "JSONBbox",
    "JSONBooleanAttribute",
    "JSONCoordinateSystem",
    "JSONCuboid",
    "JSONElementDataPointer",
    "JSONFrameInterval",
    "JSONFrame",
    "JSONFrameData",
    "JSONFrameProperties",
    "JSONMetadata",
    "JSONNumAttribute",
    "JSONNum",
    "JSONObjectData",
    "JSONObject",
    "JSONPoly2d",
    "JSONPoly3d",
    "JSONScene",
    "JSONSceneContent",
    "JSONStreamCamera",
    "JSONStreamCameraProperties",
    "JSONIntrinsicsPinhole",
    "JSONStreamOther",
    "JSONStreamRadar",
    "JSONStreamRadarProperties",
    "JSONIntrinsicsRadar",
    "JSONStreamSync",
    "JSONStreamSyncProperties",
    "JSONStreamSyncTimestamp",
    "JSONTextAttribute",
    "JSONTransformData",
    "JSONVecAttribute",
    "JSONVec",
]



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import pytest

from raillabel.json_format import (
    JSONAttributes,
    JSONBooleanAttribute,
    JSONNumAttribute,
    JSONTextAttribute,
    JSONVecAttribute,
)
from raillabel.format._attributes import (
    _attributes_from_json,
    _attributes_to_json,
    UnsupportedAttributeTypeError,
)

# == Fixtures =========================


@pytest.fixture
def attributes_multiple_types_json() -> JSONAttributes:
    return JSONAttributes(
        boolean=[
            JSONBooleanAttribute(name="has_red_hat", val=True),
            JSONBooleanAttribute(name="has_green_hat", val=False),
        ],
        num=[JSONNumAttribute(name="number_of_red_clothing_items", val=2)],
        text=[JSONTextAttribute(name="color_of_hat", val="red")],
        vec=[
            JSONVecAttribute(
                name="clothing_items", val=["red_hat", "brown_coat", "black_pants", "red_shoes"]
            )
        ],
    )


@pytest.fixture
def attributes_multiple_types() -> dict:
    return {
        "has_red_hat": True,
        "has_green_hat": False,
        "number_of_red_clothing_items": 2,
        "color_of_hat": "red",
        "clothing_items": ["red_hat", "brown_coat", "black_pants", "red_shoes"],
    }


# == Tests ============================


def test_attributes_from_json__none():
    actual = _attributes_from_json(None)
    assert actual == {}


def test_attributes_from_json__empty():
    json_attributes = JSONAttributes(
        boolean=None,
        num=None,
        text=None,
        vec=None,
    )

    actual = _attributes_from_json(json_attributes)
    assert actual == {}


def test_attributes_from_json__multiple_types(
    attributes_multiple_types, attributes_multiple_types_json
):
    actual = _attributes_from_json(attributes_multiple_types_json)
    assert actual == attributes_multiple_types


def test_attributes_to_json__empty():
    actual = _attributes_to_json({})
    assert actual == None


def test_attributes_to_json__multiple_types(
    attributes_multiple_types, attributes_multiple_types_json
):
    actual = _attributes_to_json(attributes_multiple_types)
    assert actual == attributes_multiple_types_json


def test_attributes_to_json__unsupported_type():
    with pytest.raises(UnsupportedAttributeTypeError):
        _attributes_to_json({"attribute_with_unsupported_type": object})


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import pytest

from raillabel.format import Size3d

# == Fixtures =========================


@pytest.fixture
def size3d_json() -> dict:
    return [25, 1.344, 12.3]


@pytest.fixture
def size3d() -> dict:
    return Size3d(25, 1.344, 12.3)


# == Tests ============================


def test_from_json(size3d, size3d_json):
    actual = Size3d.from_json(size3d_json)
    assert actual == size3d


def test_to_json(size3d, size3d_json):
    actual = size3d.to_json()
    assert actual == tuple(size3d_json)


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import pytest

import raillabel


def test_load(json_paths):
    actual = raillabel.load(json_paths["openlabel_v1_short"])
    assert len(actual.sensors) == 4
    assert len(actual.objects) == 3
    assert len(actual.frames) == 2


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0
"""Package for the scene building functionality."""

from .scene_builder import SceneBuilder

__all__ = ["SceneBuilder"]



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from dataclasses import dataclass, field
from decimal import Decimal
from uuid import UUID

from raillabel.json_format import (
    JSONAnnotations,
    JSONFrame,
    JSONFrameData,
    JSONFrameProperties,
    JSONObjectData,
)

from ._util import _empty_list_to_none
from .bbox import Bbox
from .cuboid import Cuboid
from .num import Num
from .object import Object
from .poly2d import Poly2d
from .poly3d import Poly3d
from .seg3d import Seg3d
from .sensor_reference import SensorReference


@dataclass
class Frame:
    """A container of dynamic, timewise, information."""

    timestamp: Decimal | None = None
    "Timestamp containing the Unix epoch time of the frame with up to nanosecond precision."

    sensors: dict[str, SensorReference] = field(default_factory=dict)
    "References to the sensors with frame specific information like timestamp and uri."

    frame_data: dict[str, Num] = field(default_factory=dict)
    """Dictionary containing data directly connected to the frame and not to any object, like
    gps/imu data. Dictionary keys are the ID-strings of the variable the data belongs to."""

    annotations: dict[UUID, Bbox | Cuboid | Poly2d | Poly3d | Seg3d] = field(default_factory=dict)
    "All annotations of this frame."

    @classmethod
    def from_json(cls, json: JSONFrame) -> Frame:
        """Construct an instant of this class from RailLabel JSON data."""
        return Frame(
            timestamp=_timestamp_from_dict(json.frame_properties),
            sensors=_sensors_from_dict(json.frame_properties),
            frame_data=_frame_data_from_dict(json.frame_properties),
            annotations=_annotations_from_json(json.objects),
        )

    def to_json(self, objects: dict[UUID, Object]) -> JSONFrame:
        """Export this object into the RailLabel JSON format."""
        return JSONFrame(
            frame_properties=JSONFrameProperties(
                timestamp=self.timestamp,
                streams={
                    sensor_id: sensor_ref.to_json() for sensor_id, sensor_ref in self.sensors.items()
                },
                frame_data=JSONFrameData(num=[num.to_json() for num in self.frame_data.values()]),
            ),
            objects=_objects_to_json(self.annotations, objects),
        )


def _timestamp_from_dict(frame_properties: JSONFrameProperties | None) -> Decimal | None:
    if frame_properties is None:
        return None

    if frame_properties.timestamp is None:
        return None

    return Decimal(frame_properties.timestamp)


def _sensors_from_dict(frame_properties: JSONFrameProperties | None) -> dict[str, SensorReference]:
    if frame_properties is None:
        return {}

    if frame_properties.streams is None:
        return {}

    return {
        sensor_id: SensorReference.from_json(sensor_ref)
        for sensor_id, sensor_ref in frame_properties.streams.items()
    }


def _frame_data_from_dict(frame_properties: JSONFrameProperties | None) -> dict[str, Num]:
    if frame_properties is None:
        return {}

    if frame_properties.frame_data is None:
        return {}

    if frame_properties.frame_data.num is None:
        return {}

    return {num.name: Num.from_json(num) for num in frame_properties.frame_data.num}


def _annotations_from_json(
    json_object_data: dict[UUID, JSONObjectData] | None,
) -> dict[UUID, Bbox | Cuboid | Poly2d | Poly3d | Seg3d]:
    if json_object_data is None:
        return {}

    annotations: dict[UUID, Bbox | Cuboid | Poly2d | Poly3d | Seg3d] = {}

    for object_id, object_data in json_object_data.items():
        for json_bbox in _resolve_none_to_empty_list(object_data.object_data.bbox):
            annotations[json_bbox.uid] = Bbox.from_json(json_bbox, object_id)

        for json_cuboid in _resolve_none_to_empty_list(object_data.object_data.cuboid):
            annotations[json_cuboid.uid] = Cuboid.from_json(json_cuboid, object_id)

        for json_poly2d in _resolve_none_to_empty_list(object_data.object_data.poly2d):
            annotations[json_poly2d.uid] = Poly2d.from_json(json_poly2d, object_id)

        for json_poly3d in _resolve_none_to_empty_list(object_data.object_data.poly3d):
            annotations[json_poly3d.uid] = Poly3d.from_json(json_poly3d, object_id)

        for json_seg3d in _resolve_none_to_empty_list(object_data.object_data.vec):
            annotations[json_seg3d.uid] = Seg3d.from_json(json_seg3d, object_id)

    return annotations


def _resolve_none_to_empty_list(optional_list: list | None) -> list:
    if optional_list is None:
        return []
    return optional_list


def _objects_to_json(
    annotations: dict[UUID, Bbox | Cuboid | Poly2d | Poly3d | Seg3d], objects: dict[UUID, Object]
) -> dict[str, JSONObjectData] | None:
    if len(annotations) == 0:
        return None

    object_data = {}

    for ann_id, annotation in annotations.items():
        object_id = str(annotation.object_id)

        if object_id not in object_data:
            object_data[object_id] = JSONObjectData(
                object_data=JSONAnnotations(
                    bbox=[],
                    cuboid=[],
                    poly2d=[],
                    poly3d=[],
                    vec=[],
                )
            )

        json_annotation = annotation.to_json(ann_id, objects[UUID(object_id)].type)

        if isinstance(annotation, Bbox):
            object_data[object_id].object_data.bbox.append(json_annotation)  # type: ignore

        elif isinstance(annotation, Cuboid):
            object_data[object_id].object_data.cuboid.append(json_annotation)  # type: ignore

        elif isinstance(annotation, Poly2d):
            object_data[object_id].object_data.poly2d.append(json_annotation)  # type: ignore

        elif isinstance(annotation, Poly3d):
            object_data[object_id].object_data.poly3d.append(json_annotation)  # type: ignore

        elif isinstance(annotation, Seg3d):
            object_data[object_id].object_data.vec.append(json_annotation)  # type: ignore

        else:
            raise TypeError

    for object_id in object_data:
        object_data[object_id].object_data.bbox = _empty_list_to_none(
            object_data[object_id].object_data.bbox
        )
        object_data[object_id].object_data.cuboid = _empty_list_to_none(
            object_data[object_id].object_data.cuboid
        )
        object_data[object_id].object_data.poly2d = _empty_list_to_none(
            object_data[object_id].object_data.poly2d
        )
        object_data[object_id].object_data.poly3d = _empty_list_to_none(
            object_data[object_id].object_data.poly3d
        )
        object_data[object_id].object_data.vec = _empty_list_to_none(
            object_data[object_id].object_data.vec
        )

    return object_data



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from dataclasses import dataclass


@dataclass
class Quaternion:
    """A rotation represented by a quaternion."""

    x: float
    "The x component of the quaternion."

    y: float
    "The y component of the quaternion."

    z: float
    "The z component of the quaternion."

    w: float
    "The omega component of the quaternion."

    @classmethod
    def from_json(cls, json: tuple[float, float, float, float]) -> Quaternion:
        """Construct an instant of this class from RailLabel JSON data."""
        return Quaternion(x=json[0], y=json[1], z=json[2], w=json[3])

    def to_json(self) -> tuple[float, float, float, float]:
        """Export this object into the RailLabel JSON format."""
        return (self.x, self.y, self.z, self.w)



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from dataclasses import dataclass

from raillabel.json_format import JSONIntrinsicsPinhole


@dataclass
class IntrinsicsPinhole:
    """Intrinsic calibration for a camera sensor."""

    camera_matrix: tuple[
        float, float, float, float, float, float, float, float, float, float, float, float
    ]
    """This is a 3x4 camera matrix which projects 3D homogeneous points (4x1) from a camera
    coordinate system into the image plane (3x1). This is the usual K matrix for camera projection as
    in OpenCV. It is extended from 3x3 to 3x4 to enable its direct utilisation to project 4x1
    homogeneous 3D points. The matrix is defined to follow the camera model: x-to-right, y-down,
    z-forward. The following equation applies: x_img = camera_matrix * X_ccs."""

    distortion: tuple[float, float, float, float, float]
    "This is the array 1x5 radial and tangential distortion coefficients."

    width_px: int
    "Width of the image frame in pixel."

    height_px: int
    "Height of the image frame in pixel."

    @classmethod
    def from_json(cls, json: JSONIntrinsicsPinhole) -> IntrinsicsPinhole:
        """Construct an instant of this class from RailLabel JSON data."""
        return IntrinsicsPinhole(
            camera_matrix=json.camera_matrix,
            distortion=json.distortion_coeffs,
            width_px=json.width_px,
            height_px=json.height_px,
        )

    def to_json(self) -> JSONIntrinsicsPinhole:
        """Export this object into the RailLabel JSON format."""
        return JSONIntrinsicsPinhole(
            camera_matrix=self.camera_matrix,
            distortion_coeffs=self.distortion,
            width_px=self.width_px,
            height_px=self.height_px,
        )



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from raillabel.json_format import JSONCoordinateSystem, JSONStreamOther

from ._sensor_without_intrinsics import _extrinsics_from_json, _SensorWithoutIntrinsics


class GpsImu(_SensorWithoutIntrinsics):
    """A gps sensor with inertial measurement unit."""

    TYPE: str = "gps_imu"

    @classmethod
    def from_json(
        cls, json_stream: JSONStreamOther, json_coordinate_system: JSONCoordinateSystem
    ) -> GpsImu:
        """Construct an instant of this class from RailLabel JSON data."""
        return GpsImu(
            extrinsics=_extrinsics_from_json(json_coordinate_system.pose_wrt_parent),
            uri=json_stream.uri,
            description=json_stream.description,
        )

    def to_json(self) -> tuple[JSONStreamOther, JSONCoordinateSystem]:
        """Export this object into the RailLabel JSON format."""
        return (
            JSONStreamOther(
                type="gps_imu",
                uri=self.uri,
                description=self.description,
            ),
            JSONCoordinateSystem(
                parent="base",
                type="sensor",
                pose_wrt_parent=self.extrinsics.to_json() if self.extrinsics is not None else None,
            ),
        )



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import pytest

from raillabel.format import Transform
from raillabel.json_format import JSONTransformData


# == Fixtures =========================


@pytest.fixture
def transform_json(point3d_json, quaternion_json) -> JSONTransformData:
    return JSONTransformData(translation=point3d_json, quaternion=quaternion_json)


@pytest.fixture
def transform(point3d, quaternion) -> Transform:
    return Transform(pos=point3d, quat=quaternion)


# == Tests ============================


def test_from_json(transform_json, transform):
    actual = Transform.from_json(transform_json)
    assert actual == transform


def test_to_json(transform, transform_json):
    actual = transform.to_json()
    assert actual == transform_json


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from uuid import UUID

import pytest

from raillabel.format import Bbox
from raillabel.json_format import JSONBbox

# == Fixtures =========================


@pytest.fixture
def bbox_json(
    attributes_multiple_types_json,
    point2d_json,
    size2d_json,
) -> JSONBbox:
    return JSONBbox(
        uid="2811f67c-124C-4fac-a275-20807d0471de",
        name="rgb_middle__bbox__person",
        val=point2d_json + size2d_json,
        coordinate_system="rgb_middle",
        attributes=attributes_multiple_types_json,
    )


@pytest.fixture
def bbox_id() -> UUID:
    return UUID("2811f67c-124C-4fac-a275-20807d0471de")


@pytest.fixture
def bbox(
    point2d,
    size2d,
    attributes_multiple_types,
    object_person_id,
) -> Bbox:
    return Bbox(
        pos=point2d,
        size=size2d,
        sensor_id="rgb_middle",
        attributes=attributes_multiple_types,
        object_id=object_person_id,
    )


# == Tests ============================


def test_from_json(bbox, bbox_json, object_person_id):
    actual = Bbox.from_json(bbox_json, object_person_id)
    assert actual == bbox


def test_name(bbox):
    actual = bbox.name("person")
    assert actual == "rgb_middle__bbox__person"


def test_to_json(bbox, bbox_json, bbox_id):
    actual = bbox.to_json(bbox_id, object_type="person")
    assert actual == bbox_json


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from dataclasses import dataclass

from raillabel.json_format import JSONMetadata


@dataclass
class Metadata:
    """Container for metadata information about the scene itself."""

    schema_version: str
    "Version number of the OpenLABEL schema this annotation object follows."

    annotator: str | None = None
    "Name or description of the annotator that created the annotations."

    comment: str | None = None
    "Additional information or description about the annotation content."

    exporter_version: str | None = None
    "Version of the raillabel-devkit, that last exported the scene."

    file_version: str | None = None
    "Version number of the raillabel annotation content."

    name: str | None = None
    "Name of the raillabel annotation content."

    subschema_version: str | None = None
    "Version number of the RailLabel schema this annotation object follows."

    tagged_file: str | None = None
    "Directory with the exported data_dict (e.g. images, point clouds)."

    @classmethod
    def from_json(cls, json: JSONMetadata) -> Metadata:
        """Construct an instant of this class from RailLabel JSON data."""
        metadata = Metadata(
            schema_version=json.schema_version,
            name=json.name,
            subschema_version=json.subschema_version,
            exporter_version=json.exporter_version,
            file_version=json.file_version,
            tagged_file=json.tagged_file,
            annotator=json.annotator,
            comment=json.comment,
        )

        if json.model_extra is not None:
            for extra_field, extra_value in json.model_extra.items():
                setattr(metadata, extra_field, extra_value)

        return metadata

    def to_json(self) -> JSONMetadata:
        """Export this object into the RailLabel JSON format."""
        return JSONMetadata(**dict(vars(self)))



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from dataclasses import dataclass

from raillabel.json_format import JSONTransformData

from .transform import Transform


@dataclass
class _SensorWithoutIntrinsics:
    """Parent class of all sensors, that do not have an intrinsic calibration."""

    extrinsics: Transform | None = None
    "External calibration of the sensor defined by the 3D transform to the coordinate system origin."

    uri: str | None = None
    "Name of the subdirectory containing the sensor files."

    description: str | None = None
    "Additional information about the sensor."


def _extrinsics_from_json(json_transform: JSONTransformData | None) -> Transform | None:
    if json_transform is None:
        return None
    return Transform.from_json(json_transform)



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from ._json_format_base import _JSONFormatBase
from .element_data_pointer import JSONElementDataPointer
from .frame_interval import JSONFrameInterval


class JSONObject(_JSONFormatBase):
    """An object is the main type of annotation element.

    Object is designed to represent spatiotemporal entities, such as physical objects in the real
    world. Objects shall have a name and type. Objects may have static and dynamic data. Objects
    are the only type of elements that may have geometric data, such as bounding boxes, cuboids,
    polylines, images, etc.
    """

    name: str
    "Name of the object. It is a friendly name and not used for indexing."

    type: str
    "The type of an object defines the class the object corresponds to."

    frame_intervals: list[JSONFrameInterval] | None = None
    "The array of frame intervals where this object exists or is defined."

    object_data_pointers: dict[str, JSONElementDataPointer] | None = None



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from typing import Literal

from ._json_format_base import _JSONFormatBase


class JSONStreamCamera(_JSONFormatBase):
    """A stream describes the source of a data sequence, usually a sensor.

    This specific object contains the intrinsics of a camera sensor.
    """

    type: Literal["camera"]
    "A string encoding the type of the stream."

    stream_properties: JSONStreamCameraProperties
    "Intrinsic calibration of the stream."

    uri: str
    "A string encoding the subdirectory containing the sensor files."

    description: str | None = None
    "Description of the stream."


class JSONStreamCameraProperties(_JSONFormatBase):
    """Intrinsic calibration of the stream."""

    intrinsics_pinhole: JSONIntrinsicsPinhole


class JSONIntrinsicsPinhole(_JSONFormatBase):
    """JSON object defining an instance of the intrinsic parameters of a pinhole camera."""

    camera_matrix: tuple[
        float, float, float, float, float, float, float, float, float, float, float, float
    ]
    """This is a 3x4 camera matrix which projects 3D homogeneous points (4x1) from a camera
    coordinate system into the image plane (3x1). This is the usual K matrix for camera projection as
    in OpenCV. It is extended from 3x3 to 3x4 to enable its direct utilisation to project 4x1
    homogeneous 3D points. The matrix is defined to follow the camera model: x-to-right, y-down,
    z-forward. The following equation applies: x_img = camera_matrix * X_ccs."""

    distortion_coeffs: tuple[float, float, float, float, float]
    "This is the array 1x5 radial and tangential distortion coefficients."

    height_px: int
    "Height of the camera output in pixel."

    width_px: int
    "Width of the camera output in pixel."



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from typing import Literal

from pydantic import BaseModel


class JSONMetadata(BaseModel, extra="allow"):
    """Metadata about the annotation file itself."""

    schema_version: Literal["1.0.0"]
    "Version number of the OpenLABEL schema this annotation file follows."

    name: str | None = None
    "Name of the OpenLABEL annotation content."

    subschema_version: str | None = None
    "Version number of the RailLabel subschema this annotation file follows."

    exporter_version: str | None = None
    "Version identifyer of the exporter software."

    file_version: str | None = None
    "Version number of the OpenLABEL annotation content."

    tagged_file: str | None = None
    "File name or URI of the data file being tagged."

    annotator: str | None = None
    "The person or organization responsible for annotating the file."

    comment: str | None = None
    "Additional information or description about the annotation content."



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from dataclasses import dataclass
from uuid import UUID

from raillabel.json_format import JSONBbox

from ._attributes import _attributes_from_json, _attributes_to_json
from .point2d import Point2d
from .size2d import Size2d


@dataclass
class Bbox:
    """A 2D bounding box in an image."""

    pos: Point2d
    "The center point of the bbox in pixels."

    size: Size2d
    "The dimensions of the bbox in pixels from the top left corner to the bottom right corner."

    object_id: UUID
    "The unique identifyer of the real-life object, this annotation belongs to."

    sensor_id: str
    "The unique identifyer of the sensor this annotation is labeled in."

    attributes: dict[str, float | bool | str | list]
    "Additional information associated with the annotation."

    @classmethod
    def from_json(cls, json: JSONBbox, object_id: UUID) -> Bbox:
        """Construct an instant of this class from RailLabel JSON data."""
        return Bbox(
            pos=Point2d.from_json((json.val[0], json.val[1])),
            size=Size2d.from_json((json.val[2], json.val[3])),
            object_id=object_id,
            sensor_id=json.coordinate_system,
            attributes=_attributes_from_json(json.attributes),
        )

    def to_json(self, uid: UUID, object_type: str) -> JSONBbox:
        """Export this object into the RailLabel JSON format."""
        return JSONBbox(
            name=self.name(object_type),
            val=list(self.pos.to_json()) + list(self.size.to_json()),
            coordinate_system=self.sensor_id,
            uid=uid,
            attributes=_attributes_to_json(self.attributes),
        )

    def name(self, object_type: str) -> str:
        """Return the name of the annotation used for indexing in the object data pointers."""
        return f"{self.sensor_id}__bbox__{object_type}"



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import pytest

from raillabel.format import FrameInterval
from raillabel.json_format import JSONFrameInterval

# == Fixtures =========================


@pytest.fixture
def frame_interval_json() -> JSONFrameInterval:
    return JSONFrameInterval(frame_start=12, frame_end=16)


@pytest.fixture
def frame_interval() -> dict:
    return FrameInterval(
        start=12,
        end=16,
    )


# == Tests ============================


def test_from_json(frame_interval, frame_interval_json):
    actual = FrameInterval.from_json(frame_interval_json)
    assert actual == frame_interval


def test_len():
    frame_interval = FrameInterval(
        start=12,
        end=16,
    )

    assert len(frame_interval) == 5


def test_from_frame_ids_empty():
    frame_ids = []

    assert FrameInterval.from_frame_ids(frame_ids) == []


def test_from_frame_ids_one_frame():
    frame_ids = [1]

    assert FrameInterval.from_frame_ids(frame_ids) == [FrameInterval(1, 1)]


def test_from_frame_ids_one_interval():
    frame_ids = [1, 2, 3, 4]

    assert FrameInterval.from_frame_ids(frame_ids) == [FrameInterval(1, 4)]


def test_from_frame_ids_multiple_intervals():
    frame_ids = [0, 1, 2, 3, 6, 7, 9, 12, 13, 14]

    assert FrameInterval.from_frame_ids(frame_ids) == [
        FrameInterval(0, 3),
        FrameInterval(6, 7),
        FrameInterval(9, 9),
        FrameInterval(12, 14),
    ]


def test_from_frame_ids_unsorted():
    frame_ids = [5, 2, 1, 3]

    assert FrameInterval.from_frame_ids(frame_ids) == [
        FrameInterval(1, 3),
        FrameInterval(5, 5),
    ]


def test_to_json(frame_interval, frame_interval_json):
    actual = frame_interval.to_json()
    assert actual == frame_interval_json


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from dataclasses import dataclass
from decimal import Decimal

from raillabel.json_format import JSONStreamSync, JSONStreamSyncProperties, JSONStreamSyncTimestamp


@dataclass
class SensorReference:
    """A reference to a sensor in a specific frame."""

    timestamp: Decimal
    """Timestamp containing the Unix epoch time of the sensor in a specific frame with up to
    nanosecond precision."""

    uri: str | None = None
    "URI to the file corresponding to the frame recording in the particular frame."

    @classmethod
    def from_json(cls, json: JSONStreamSync) -> SensorReference:
        """Construct an instant of this class from RailLabel JSON data."""
        return SensorReference(
            timestamp=Decimal(json.stream_properties.sync.timestamp),
            uri=json.uri,
        )

    def to_json(self) -> JSONStreamSync:
        """Export this object into the RailLabel JSON format."""
        return JSONStreamSync(
            stream_properties=JSONStreamSyncProperties(
                sync=JSONStreamSyncTimestamp(timestamp=str(self.timestamp)),
            ),
            uri=self.uri,
        )



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from uuid import UUID

import pytest

from raillabel.format import Cuboid
from raillabel.json_format import JSONCuboid

# == Fixtures =========================


@pytest.fixture
def cuboid_json(
    attributes_multiple_types_json,
    point3d_json,
    quaternion_json,
    size3d_json,
) -> JSONCuboid:
    return JSONCuboid(
        uid="51def938-20BA-4699-95be-d6330c44cb77",
        name="lidar__cuboid__person",
        val=point3d_json + quaternion_json + size3d_json,
        coordinate_system="lidar",
        attributes=attributes_multiple_types_json,
    )


@pytest.fixture
def cuboid_id() -> UUID:
    return UUID("51def938-20BA-4699-95be-d6330c44cb77")


@pytest.fixture
def cuboid(
    point3d,
    size3d,
    quaternion,
    attributes_multiple_types,
    object_person_id,
) -> Cuboid:
    return Cuboid(
        pos=point3d,
        quat=quaternion,
        size=size3d,
        sensor_id="lidar",
        attributes=attributes_multiple_types,
        object_id=object_person_id,
    )


# == Tests ============================


def test_from_json(cuboid, cuboid_json, object_person_id):
    actual = Cuboid.from_json(cuboid_json, object_person_id)
    assert actual == cuboid


def test_name(cuboid):
    actual = cuboid.name("person")
    assert actual == "lidar__cuboid__person"


def test_to_json(cuboid, cuboid_json, cuboid_id):
    actual = cuboid.to_json(cuboid_id, object_type="person")
    assert actual == cuboid_json


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from dataclasses import dataclass

from raillabel.json_format import JSONFrameInterval


@dataclass
class FrameInterval:
    """Closed interval of frames."""

    start: int
    "Initial frame number of the interval (inclusive)."

    end: int
    "Ending frame number of the interval (inclusive)."

    @classmethod
    def from_json(cls, json: JSONFrameInterval) -> FrameInterval:
        """Construct an instant of this class from RailLabel JSON data."""
        return FrameInterval(
            start=json.frame_start,
            end=json.frame_end,
        )

    @classmethod
    def from_frame_ids(cls, frame_ids: list[int]) -> list[FrameInterval]:
        """Convert a list of frame uids into FrameIntervals.

        Example:
        -------
        ```python
        FrameInterval.from_frame_ids([0, 1, 2, 3, 9, 12, 13, 14]) == [
            FrameInterval(0, 3),
            FrameInterval(9, 9),
            FrameInterval(12, 14),
        ]
        ```

        """
        sorted_frame_ids = sorted(frame_ids)
        frame_id_intervals = _slice_into_intervals(sorted_frame_ids)

        return [
            FrameInterval(start=interval[0], end=interval[-1]) for interval in frame_id_intervals
        ]

    def to_json(self) -> JSONFrameInterval:
        """Export this object into the RailLabel JSON format."""
        return JSONFrameInterval(
            frame_start=self.start,
            frame_end=self.end,
        )

    def __len__(self) -> int:
        """Return the length in frames."""
        return abs(self.start - self.end) + 1


def _slice_into_intervals(sorted_frame_ids: list[int]) -> list[list[int]]:
    if len(sorted_frame_ids) == 0:
        return []

    if len(sorted_frame_ids) == 1:
        return [sorted_frame_ids]

    intervals = []
    interval_start_i = 0
    for i, frame_id in enumerate(sorted_frame_ids[1:]):
        previous_frame_id = sorted_frame_ids[i]

        if frame_id - previous_frame_id > 1:
            intervals.append(sorted_frame_ids[interval_start_i : i + 1])
            interval_start_i = i + 1

    intervals.append(sorted_frame_ids[interval_start_i : len(sorted_frame_ids)])
    interval_start_i = len(sorted_frame_ids)

    return intervals



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from dataclasses import dataclass


@dataclass
class Size3d:
    """The 3D size of a cube."""

    x: float
    "The size along the x-axis."

    y: float
    "The size along the y-axis."

    z: float
    "The size along the z-axis."

    @classmethod
    def from_json(cls, json: tuple[float, float, float]) -> Size3d:
        """Construct an instant of this class from RailLabel JSON data."""
        return Size3d(x=json[0], y=json[1], z=json[2])

    def to_json(self) -> tuple[float, float, float]:
        """Export this object into the RailLabel JSON format."""
        return (self.x, self.y, self.z)



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from dataclasses import dataclass
from uuid import UUID

from raillabel.json_format import JSONPoly2d

from ._attributes import _attributes_from_json, _attributes_to_json
from ._util import _flatten_list
from .point2d import Point2d


@dataclass
class Poly2d:
    """Sequence of 2D points. Can either be a polygon or polyline."""

    points: list[Point2d]
    "List of the 2d points that make up the polyline."

    closed: bool
    "If True, this object represents a polygon and if False, it represents a polyline."

    object_id: UUID
    "The unique identifyer of the real-life object, this annotation belongs to."

    sensor_id: str
    "The unique identifyer of the sensor this annotation is labeled in."

    attributes: dict[str, float | bool | str | list]
    "Additional information associated with the annotation."

    @classmethod
    def from_json(cls, json: JSONPoly2d, object_id: UUID) -> Poly2d:
        """Construct an instant of this class from RailLabel JSON data."""
        return Poly2d(
            points=[
                Point2d(x=float(json.val[i]), y=float(json.val[i + 1]))
                for i in range(0, len(json.val), 2)
            ],
            closed=json.closed,
            object_id=object_id,
            sensor_id=json.coordinate_system,
            attributes=_attributes_from_json(json.attributes),
        )

    def to_json(self, uid: UUID, object_type: str) -> JSONPoly2d:
        """Export this object into the RailLabel JSON format."""
        return JSONPoly2d(
            name=self.name(object_type),
            val=_flatten_list([point.to_json() for point in self.points]),
            closed=self.closed,
            mode="MODE_POLY2D_ABSOLUTE",
            coordinate_system=self.sensor_id,
            uid=uid,
            attributes=_attributes_to_json(self.attributes),
        )

    def name(self, object_type: str) -> str:
        """Return the name of the annotation used for indexing in the object data pointers."""
        return f"{self.sensor_id}__poly2d__{object_type}"



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations
from uuid import UUID
from decimal import Decimal

import pytest

import raillabel
from raillabel.scene_builder.scene_builder import SceneBuilder
from raillabel.format import (
    Scene,
    Metadata,
    Object,
    Camera,
    Lidar,
    Radar,
    GpsImu,
    OtherSensor,
    IntrinsicsPinhole,
    IntrinsicsRadar,
    Frame,
    Bbox,
    Point2d,
    Size2d,
    Cuboid,
    Point3d,
    Size3d,
    Quaternion,
    Poly2d,
    Poly3d,
    Seg3d,
)


def test_empty():
    actual = SceneBuilder.empty()
    assert actual.result == Scene(Metadata(schema_version="1.0.0"))


def test_add_object__all_options():
    actual = SceneBuilder.empty().add_object(
        object_id="5c59aad4-9fcd-4903-a9fa-72b1b76c23a5",
        object_type="train",
        object_name="train_0001",
    )
    assert actual.result == Scene(
        metadata=Metadata(schema_version="1.0.0"),
        objects={
            UUID("5c59aad4-9fcd-4903-a9fa-72b1b76c23a5"): Object(name="train_0001", type="train")
        },
    )


def test_add_object__no_object_name():
    actual = SceneBuilder.empty().add_object(
        object_id="5c59aad4-9fcd-4903-a9fa-72b1b76c23a5",
        object_type="train",
    )
    assert actual.result == Scene(
        metadata=Metadata(schema_version="1.0.0"),
        objects={
            UUID("5c59aad4-9fcd-4903-a9fa-72b1b76c23a5"): Object(name="train_0000", type="train")
        },
    )


def test_add_object__no_object_type():
    actual = SceneBuilder.empty().add_object(
        object_id="5c59aad4-9fcd-4903-a9fa-72b1b76c23a5",
        object_name="train_0000",
    )
    assert actual.result == Scene(
        metadata=Metadata(schema_version="1.0.0"),
        objects={
            UUID("5c59aad4-9fcd-4903-a9fa-72b1b76c23a5"): Object(name="train_0000", type="train")
        },
    )


def test_add_object__no_object_type_or_name():
    actual = SceneBuilder.empty().add_object(
        object_id="5c59aad4-9fcd-4903-a9fa-72b1b76c23a5",
    )
    assert actual.result == Scene(
        metadata=Metadata(schema_version="1.0.0"),
        objects={
            UUID("5c59aad4-9fcd-4903-a9fa-72b1b76c23a5"): Object(name="person_0000", type="person")
        },
    )


def test_add_object__no_object_id():
    actual = SceneBuilder.empty().add_object(object_type="train", object_name="train_0001")
    assert actual.result == Scene(
        metadata=Metadata(schema_version="1.0.0"),
        objects={
            UUID("5c59aad4-0000-4000-0000-000000000000"): Object(name="train_0001", type="train")
        },
    )


def test_add_object__object_id_iteration():
    actual = SceneBuilder.empty().add_object().add_object().add_object().add_object()
    assert actual.result == Scene(
        metadata=Metadata(schema_version="1.0.0"),
        objects={
            UUID("5c59aad4-0000-4000-0000-000000000000"): Object(name="person_0000", type="person"),
            UUID("5c59aad4-0000-4000-0000-000000000001"): Object(name="person_0000", type="person"),
            UUID("5c59aad4-0000-4000-0000-000000000002"): Object(name="person_0000", type="person"),
            UUID("5c59aad4-0000-4000-0000-000000000003"): Object(name="person_0000", type="person"),
        },
    )


def test_add_sensor__camera_rgb(camera_empty):
    actual = SceneBuilder.empty().add_sensor("rgb_middle")
    assert actual.result == Scene(
        metadata=Metadata(schema_version="1.0.0"),
        sensors={"rgb_middle": camera_empty},
    )


def test_add_sensor__camera_ir(camera_empty):
    actual = SceneBuilder.empty().add_sensor("ir_left")
    assert actual.result == Scene(
        metadata=Metadata(schema_version="1.0.0"),
        sensors={"ir_left": camera_empty},
    )


def test_add_sensor__radar(radar_empty):
    actual = SceneBuilder.empty().add_sensor("radar")
    assert actual.result == Scene(
        metadata=Metadata(schema_version="1.0.0"),
        sensors={"radar": radar_empty},
    )


def test_add_sensor__lidar():
    actual = SceneBuilder.empty().add_sensor("lidar")
    assert actual.result == Scene(
        metadata=Metadata(schema_version="1.0.0"), sensors={"lidar": Lidar()}
    )


def test_add_sensor__gps_imu():
    actual = SceneBuilder.empty().add_sensor("gps_imu")
    assert actual.result == Scene(
        metadata=Metadata(schema_version="1.0.0"), sensors={"gps_imu": GpsImu()}
    )


def test_add_sensor__other():
    actual = SceneBuilder.empty().add_sensor("SOMETHING_ELSE")
    assert actual.result == Scene(
        metadata=Metadata(schema_version="1.0.0"), sensors={"SOMETHING_ELSE": OtherSensor()}
    )


def test_add_frame():
    actual = SceneBuilder.empty().add_frame(1, 1631691173)
    assert actual.result == Scene(
        metadata=Metadata(schema_version="1.0.0"), frames={1: Frame(timestamp=Decimal(1631691173))}
    )


def test_add_frame__no_timestamp():
    actual = SceneBuilder.empty().add_frame(1)
    assert actual.result == Scene(
        metadata=Metadata(schema_version="1.0.0"), frames={1: Frame(timestamp=None)}
    )


def test_add_frame__no_frame_id():
    actual = SceneBuilder.empty().add_frame().add_frame()
    assert actual.result == Scene(
        metadata=Metadata(schema_version="1.0.0"),
        frames={
            1: Frame(),
            2: Frame(),
        },
    )


def test_add_bbox(camera_empty):
    actual = SceneBuilder.empty().add_bbox(
        uid=UUID("6c95543d-4d4f-43df-a52d-36bf868e09d8"),
        frame_id=2,
        object_name="person_0001",
        sensor_id="ir_middle",
        attributes={"attr": True},
    )
    assert actual.result == Scene(
        metadata=Metadata(schema_version="1.0.0"),
        objects={
            UUID("5c59aad4-0000-4000-0000-000000000000"): Object(name="person_0001", type="person")
        },
        sensors={"ir_middle": camera_empty},
        frames={
            2: Frame(
                annotations={
                    UUID("6c95543d-4d4f-43df-a52d-36bf868e09d8"): Bbox(
                        sensor_id="ir_middle",
                        object_id=UUID("5c59aad4-0000-4000-0000-000000000000"),
                        pos=Point2d(0, 0),
                        size=Size2d(0, 0),
                        attributes={"attr": True},
                    )
                }
            ),
        },
    )


def test_add_bbox__just_defaults(camera_empty):
    actual = SceneBuilder.empty().add_bbox()
    assert actual.result == Scene(
        metadata=Metadata(schema_version="1.0.0"),
        objects={
            UUID("5c59aad4-0000-4000-0000-000000000000"): Object(name="person_0001", type="person")
        },
        sensors={"rgb_middle": camera_empty},
        frames={
            1: Frame(
                annotations={
                    UUID("6c95543d-0000-4000-0000-000000000000"): Bbox(
                        sensor_id="rgb_middle",
                        object_id=UUID("5c59aad4-0000-4000-0000-000000000000"),
                        pos=Point2d(0, 0),
                        size=Size2d(0, 0),
                        attributes={},
                    )
                }
            ),
        },
    )


def test_add_cuboid():
    actual = SceneBuilder.empty().add_cuboid(
        uid=UUID("6c95543d-4d4f-43df-a52d-36bf868e09d8"),
        frame_id=2,
        object_name="person_0001",
        sensor_id="lidar_left",
        attributes={"my_attr": 5},
    )
    assert actual.result == Scene(
        metadata=Metadata(schema_version="1.0.0"),
        objects={
            UUID("5c59aad4-0000-4000-0000-000000000000"): Object(name="person_0001", type="person")
        },
        sensors={"lidar_left": Lidar()},
        frames={
            2: Frame(
                annotations={
                    UUID("6c95543d-4d4f-43df-a52d-36bf868e09d8"): Cuboid(
                        sensor_id="lidar_left",
                        object_id=UUID("5c59aad4-0000-4000-0000-000000000000"),
                        pos=Point3d(0, 0, 0),
                        size=Size3d(0, 0, 0),
                        quat=Quaternion(0, 0, 0, 0),
                        attributes={"my_attr": 5},
                    )
                }
            ),
        },
    )


def test_add_cuboid__just_defaults():
    actual = SceneBuilder.empty().add_cuboid()
    assert actual.result == Scene(
        metadata=Metadata(schema_version="1.0.0"),
        objects={
            UUID("5c59aad4-0000-4000-0000-000000000000"): Object(name="person_0001", type="person")
        },
        sensors={"lidar": Lidar()},
        frames={
            1: Frame(
                annotations={
                    UUID("6c95543d-0000-4000-0000-000000000000"): Cuboid(
                        sensor_id="lidar",
                        object_id=UUID("5c59aad4-0000-4000-0000-000000000000"),
                        pos=Point3d(0, 0, 0),
                        size=Size3d(0, 0, 0),
                        quat=Quaternion(0, 0, 0, 0),
                        attributes={},
                    )
                }
            ),
        },
    )


def test_add_poly2d(camera_empty):
    actual = SceneBuilder.empty().add_poly2d(
        uid=UUID("6c95543d-4d4f-43df-a52d-36bf868e09d8"),
        frame_id=2,
        object_name="person_0001",
        sensor_id="ir_left",
        attributes={"my_attr": 5},
    )
    assert actual.result == Scene(
        metadata=Metadata(schema_version="1.0.0"),
        objects={
            UUID("5c59aad4-0000-4000-0000-000000000000"): Object(name="person_0001", type="person")
        },
        sensors={"ir_left": camera_empty},
        frames={
            2: Frame(
                annotations={
                    UUID("6c95543d-4d4f-43df-a52d-36bf868e09d8"): Poly2d(
                        sensor_id="ir_left",
                        object_id=UUID("5c59aad4-0000-4000-0000-000000000000"),
                        points=[],
                        closed=False,
                        attributes={"my_attr": 5},
                    )
                }
            ),
        },
    )


def test_add_poly2d__just_defaults(camera_empty):
    actual = SceneBuilder.empty().add_poly2d()
    assert actual.result == Scene(
        metadata=Metadata(schema_version="1.0.0"),
        objects={
            UUID("5c59aad4-0000-4000-0000-000000000000"): Object(name="person_0001", type="person")
        },
        sensors={"rgb_middle": camera_empty},
        frames={
            1: Frame(
                annotations={
                    UUID("6c95543d-0000-4000-0000-000000000000"): Poly2d(
                        sensor_id="rgb_middle",
                        object_id=UUID("5c59aad4-0000-4000-0000-000000000000"),
                        points=[],
                        closed=False,
                        attributes={},
                    )
                }
            ),
        },
    )


def test_add_poly3d():
    actual = SceneBuilder.empty().add_poly3d(
        uid=UUID("6c95543d-4d4f-43df-a52d-36bf868e09d8"),
        frame_id=2,
        object_name="person_0001",
        sensor_id="lidar_right",
        attributes={"my_attr": 5},
    )
    assert actual.result == Scene(
        metadata=Metadata(schema_version="1.0.0"),
        objects={
            UUID("5c59aad4-0000-4000-0000-000000000000"): Object(name="person_0001", type="person")
        },
        sensors={"lidar_right": Lidar()},
        frames={
            2: Frame(
                annotations={
                    UUID("6c95543d-4d4f-43df-a52d-36bf868e09d8"): Poly3d(
                        sensor_id="lidar_right",
                        object_id=UUID("5c59aad4-0000-4000-0000-000000000000"),
                        points=[],
                        closed=False,
                        attributes={"my_attr": 5},
                    )
                }
            ),
        },
    )


def test_add_poly3d__just_defaults():
    actual = SceneBuilder.empty().add_poly3d()
    assert actual.result == Scene(
        metadata=Metadata(schema_version="1.0.0"),
        objects={
            UUID("5c59aad4-0000-4000-0000-000000000000"): Object(name="person_0001", type="person")
        },
        sensors={"lidar": Lidar()},
        frames={
            1: Frame(
                annotations={
                    UUID("6c95543d-0000-4000-0000-000000000000"): Poly3d(
                        sensor_id="lidar",
                        object_id=UUID("5c59aad4-0000-4000-0000-000000000000"),
                        points=[],
                        closed=False,
                        attributes={},
                    )
                }
            ),
        },
    )


def test_add_seg3d():
    actual = SceneBuilder.empty().add_seg3d(
        uid=UUID("6c95543d-4d4f-43df-a52d-36bf868e09d8"),
        frame_id=2,
        object_name="person_0001",
        sensor_id="lidar_right",
        attributes={"my_attr": 5},
    )
    assert actual.result == Scene(
        metadata=Metadata(schema_version="1.0.0"),
        objects={
            UUID("5c59aad4-0000-4000-0000-000000000000"): Object(name="person_0001", type="person")
        },
        sensors={"lidar_right": Lidar()},
        frames={
            2: Frame(
                annotations={
                    UUID("6c95543d-4d4f-43df-a52d-36bf868e09d8"): Seg3d(
                        sensor_id="lidar_right",
                        object_id=UUID("5c59aad4-0000-4000-0000-000000000000"),
                        point_ids=[],
                        attributes={"my_attr": 5},
                    )
                }
            ),
        },
    )


def test_add_seg3d__just_defaults():
    actual = SceneBuilder.empty().add_seg3d()
    assert actual.result == Scene(
        metadata=Metadata(schema_version="1.0.0"),
        objects={
            UUID("5c59aad4-0000-4000-0000-000000000000"): Object(name="person_0001", type="person")
        },
        sensors={"lidar": Lidar()},
        frames={
            1: Frame(
                annotations={
                    UUID("6c95543d-0000-4000-0000-000000000000"): Seg3d(
                        sensor_id="lidar",
                        object_id=UUID("5c59aad4-0000-4000-0000-000000000000"),
                        point_ids=[],
                        attributes={},
                    )
                }
            ),
        },
    )


if __name__ == "__main__":
    pytest.main([__file__, "-vv"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from uuid import UUID

import pytest

from raillabel.format import Seg3d
from raillabel.json_format import JSONVec

# == Fixtures =========================


@pytest.fixture
def seg3d_json(
    attributes_multiple_types_json,
) -> JSONVec:
    return JSONVec(
        uid="d52e2b25-0B48-4899-86d5-4bc41be6b7d3",
        name="lidar__vec__person",
        val=[1234, 5678],
        coordinate_system="lidar",
        attributes=attributes_multiple_types_json,
    )


@pytest.fixture
def seg3d_id() -> UUID:
    return UUID("d52e2b25-0B48-4899-86d5-4bc41be6b7d3")


@pytest.fixture
def seg3d(
    attributes_multiple_types,
    object_person_id,
) -> Seg3d:
    return Seg3d(
        point_ids=[1234, 5678],
        sensor_id="lidar",
        attributes=attributes_multiple_types,
        object_id=object_person_id,
    )


# == Tests ============================


def test_from_json(seg3d, seg3d_json, object_person_id):
    actual = Seg3d.from_json(seg3d_json, object_person_id)
    assert actual == seg3d


def test_name(seg3d):
    actual = seg3d.name("person")
    assert actual == "lidar__vec__person"


def test_to_json(seg3d, seg3d_json, seg3d_id):
    actual = seg3d.to_json(seg3d_id, object_type="person")
    assert actual == seg3d_json


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from uuid import UUID

from ._json_format_base import _JSONFormatBase


class JSONNum(_JSONFormatBase):
    """A number."""

    name: str
    """This is a string encoding the name of this object data. It is used as index inside the
    corresponding object data pointers."""

    val: float
    "The numerical value of the number."

    coordinate_system: str | None = None
    "Name of the coordinate system in respect of which this object data is expressed."

    uid: UUID | None = None
    "This is a string encoding the Universal Unique identifyer of the annotation."



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from uuid import UUID

from ._json_format_base import _JSONFormatBase
from .attributes import JSONAttributes


class JSONCuboid(_JSONFormatBase):
    """A cuboid or 3D bounding box.

    It is defined by the position of its center, the rotation in 3D, and its dimensions.
    """

    name: str
    """This is a string encoding the name of this object data. It is used as index inside the
    corresponding object data pointers."""

    val: tuple[float, float, float, float, float, float, float, float, float, float]
    """List of values encoding the position, rotation and dimensions. It is
    (x, y, z, qx, qy, qz, qw, sx, sy, sz) where (x, y, z) encodes the position, (qx, qy, qz, qw)
    encodes the quaternion that encode the rotation, and (sx, sy, sz) are the dimensions of the
    cuboid in its object coordinate system"""

    coordinate_system: str
    "Name of the coordinate system in respect of which this object data is expressed."

    uid: UUID | None = None
    "This is a string encoding the Universal Unique identifyer of the annotation."

    attributes: JSONAttributes | None = None



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from copy import deepcopy
from dataclasses import dataclass
from decimal import Decimal
from uuid import UUID

from raillabel.format import (
    Bbox,
    Camera,
    Cuboid,
    Frame,
    GpsImu,
    IntrinsicsPinhole,
    IntrinsicsRadar,
    Lidar,
    Metadata,
    Object,
    OtherSensor,
    Point2d,
    Point3d,
    Poly2d,
    Poly3d,
    Quaternion,
    Radar,
    Scene,
    Seg3d,
    Size2d,
    Size3d,
)
from raillabel.format._util import _flatten_list


@dataclass
class SceneBuilder:
    """Use this class for easily creating scenes for tests."""

    result: Scene

    @classmethod
    def empty(cls) -> SceneBuilder:
        """Construct the SceneBuilder with an empty scene."""
        return SceneBuilder(Scene(metadata=Metadata(schema_version="1.0.0")))

    def add_object(
        self,
        object_id: str | UUID | None = None,
        object_type: str | None = None,
        object_name: str | None = None,
    ) -> SceneBuilder:
        """Add an object to the scene."""
        scene = deepcopy(self.result)

        object_type, object_name = _resolve_empty_object_name_or_type(object_type, object_name)
        object_id = _resolve_empty_object_uid(scene, object_id)

        scene.objects[object_id] = Object(object_name, object_type)
        return SceneBuilder(scene)

    def add_sensor(self, sensor_id: str) -> SceneBuilder:
        """Add a sensor to the scene.

        The sensor type is implicitly determined by the sensor_id. Id's, that start with 'rgb_' or
        'ir_' are added as a camera. If the id is 'lidar', it is a lidar. 'radar' creates a Radar.
        'gps_imu' creates a GpsImu. If the id does not match any of these, a OtherSensor is added.
        """
        scene = deepcopy(self.result)

        truncated_sensor_id = sensor_id.split("_")[0].lower()

        if truncated_sensor_id in ["rgb", "ir"]:
            scene.sensors[sensor_id] = Camera(
                intrinsics=IntrinsicsPinhole(
                    camera_matrix=(0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0),
                    distortion=(0, 0, 0, 0, 0),
                    width_px=0,
                    height_px=0,
                )
            )

        elif truncated_sensor_id == "radar":
            scene.sensors[sensor_id] = Radar(
                intrinsics=IntrinsicsRadar(resolution_px_per_m=0, width_px=0, height_px=0)
            )

        elif truncated_sensor_id == "lidar":
            scene.sensors[sensor_id] = Lidar()

        elif truncated_sensor_id == "gps":
            scene.sensors[sensor_id] = GpsImu()

        else:
            scene.sensors[sensor_id] = OtherSensor()

        return SceneBuilder(scene)

    def add_frame(self, frame_id: int | None = None, timestamp: float | None = None) -> SceneBuilder:
        """Add a frame to the scene.

        If no frame_id is provided, the frame_id is the lowest number not currently occupied by
        another frame_id.

        Example:
        ```python
        scene = SceneBuilder.empty().add_frame(frame_id=1).add_frame(frame_id=3).add_frame().result
        assert sorted(list(scene.frames.keys())) == [1, 2, 3]
        ```
        """
        scene = deepcopy(self.result)

        if frame_id is None:
            frame_id = 1
            while frame_id in scene.frames:
                frame_id += 1

        scene.frames[frame_id] = Frame(
            timestamp=Decimal(timestamp) if timestamp is not None else None
        )

        return SceneBuilder(scene)

    def add_annotation(
        self,
        annotation: Bbox | Cuboid | Poly2d | Poly3d | Seg3d,
        uid: str | UUID | None = None,
        frame_id: int = 1,
        object_name: str = "person_0001",
        sensor_id: str = "rgb_middle",
    ) -> SceneBuilder:
        """Add an annotation to the scene."""
        new_builder = deepcopy(self)

        uid = _resolve_empty_annotation_uid(new_builder.result, uid)

        if not _is_object_name_object(object_name, new_builder.result):
            new_builder = new_builder.add_object(object_name=object_name)

        if sensor_id not in new_builder.result.sensors:
            new_builder = new_builder.add_sensor(sensor_id)

        if frame_id not in new_builder.result.frames:
            new_builder = new_builder.add_frame(frame_id=frame_id)

        annotation.object_id = _get_object_uid_from_name(object_name, new_builder.result)
        annotation.sensor_id = sensor_id

        new_builder.result.frames[frame_id].annotations[uid] = annotation
        return new_builder

    def add_bbox(
        self,
        uid: str | UUID | None = None,
        frame_id: int = 1,
        object_name: str = "person_0001",
        sensor_id: str = "rgb_middle",
        attributes: dict | None = None,
    ) -> SceneBuilder:
        """Add a bbox to the scene."""
        bbox = Bbox(
            object_id=UUID("ffffffff-ffff-4fff-ffff-ffffffffffff"),
            sensor_id=sensor_id,
            pos=Point2d(0, 0),
            size=Size2d(0, 0),
            attributes=attributes if attributes is not None else {},
        )
        return self.add_annotation(
            annotation=bbox,
            uid=uid,
            frame_id=frame_id,
            object_name=object_name,
            sensor_id=sensor_id,
        )

    def add_cuboid(
        self,
        uid: str | UUID | None = None,
        frame_id: int = 1,
        object_name: str = "person_0001",
        sensor_id: str = "lidar",
        attributes: dict | None = None,
    ) -> SceneBuilder:
        """Add a cuboid to the scene."""
        cuboid = Cuboid(
            object_id=UUID("ffffffff-ffff-4fff-ffff-ffffffffffff"),
            sensor_id=sensor_id,
            pos=Point3d(0, 0, 0),
            size=Size3d(0, 0, 0),
            quat=Quaternion(0, 0, 0, 0),
            attributes=attributes if attributes is not None else {},
        )
        return self.add_annotation(
            annotation=cuboid,
            uid=uid,
            frame_id=frame_id,
            object_name=object_name,
            sensor_id=sensor_id,
        )

    def add_poly2d(
        self,
        uid: str | UUID | None = None,
        frame_id: int = 1,
        object_name: str = "person_0001",
        sensor_id: str = "rgb_middle",
        attributes: dict | None = None,
    ) -> SceneBuilder:
        """Add a poly2d to the scene."""
        poly2d = Poly2d(
            object_id=UUID("ffffffff-ffff-4fff-ffff-ffffffffffff"),
            sensor_id=sensor_id,
            points=[],
            closed=False,
            attributes=attributes if attributes is not None else {},
        )
        return self.add_annotation(
            annotation=poly2d,
            uid=uid,
            frame_id=frame_id,
            object_name=object_name,
            sensor_id=sensor_id,
        )

    def add_poly3d(
        self,
        uid: str | UUID | None = None,
        frame_id: int = 1,
        object_name: str = "person_0001",
        sensor_id: str = "lidar",
        attributes: dict | None = None,
    ) -> SceneBuilder:
        """Add a poly3d to the scene."""
        poly3d = Poly3d(
            object_id=UUID("ffffffff-ffff-4fff-ffff-ffffffffffff"),
            sensor_id=sensor_id,
            points=[],
            closed=False,
            attributes=attributes if attributes is not None else {},
        )
        return self.add_annotation(
            annotation=poly3d,
            uid=uid,
            frame_id=frame_id,
            object_name=object_name,
            sensor_id=sensor_id,
        )

    def add_seg3d(
        self,
        uid: str | UUID | None = None,
        frame_id: int = 1,
        object_name: str = "person_0001",
        sensor_id: str = "lidar",
        attributes: dict | None = None,
    ) -> SceneBuilder:
        """Add a poly3d to the scene."""
        seg3d = Seg3d(
            object_id=UUID("ffffffff-ffff-4fff-ffff-ffffffffffff"),
            sensor_id=sensor_id,
            point_ids=[],
            attributes=attributes if attributes is not None else {},
        )
        return self.add_annotation(
            annotation=seg3d,
            uid=uid,
            frame_id=frame_id,
            object_name=object_name,
            sensor_id=sensor_id,
        )


def _resolve_empty_object_name_or_type(
    object_type: str | None, object_name: str | None
) -> tuple[str, str]:
    if object_name is None and object_type is None:
        object_type = "person"
        object_name = object_type + "_0000"
        return object_type, object_name

    if object_name is None and object_type is not None:
        object_name = object_type + "_0000"
        return object_type, object_name

    if object_name is not None and object_type is None:
        object_type = object_name.split("_")[0]
        return object_type, object_name

    if object_name is not None and object_type is not None:
        return object_type, object_name

    raise RuntimeError  # this part is unreachable but this is the only way that mypy is happy


def _resolve_empty_object_uid(scene: Scene, object_id: str | UUID | None) -> UUID:
    if object_id is None:
        uid_index = 0
        while _generate_deterministic_uuid(uid_index, "5c59aad4") in scene.objects:
            uid_index += 1

        object_id = _generate_deterministic_uuid(uid_index, "5c59aad4")

    return UUID(str(object_id))


def _generate_deterministic_uuid(index: int, prefix: str) -> UUID:
    return UUID(f"{prefix.zfill(0)}-0000-4000-0000-{str(index).zfill(12)}")


def _resolve_empty_annotation_uid(scene: Scene, uid: str | UUID | None) -> UUID:
    if uid is None:
        annotation_uids_in_scene = set(
            _flatten_list([tuple(frame.annotations.keys()) for frame in scene.frames.values()])
        )
        uid_index = 0
        while _generate_deterministic_uuid(uid_index, "6c95543d") in annotation_uids_in_scene:
            uid_index += 1

        uid = _generate_deterministic_uuid(uid_index, "6c95543d")

    return UUID(str(uid))


def _is_object_name_object(object_name: str, scene: Scene) -> bool:
    return any(object_.name == object_name for object_ in scene.objects.values())


def _get_object_uid_from_name(object_name: str, scene: Scene) -> UUID:
    for object_uid, object_ in scene.objects.items():
        if object_.name == object_name:
            return object_uid

    raise RuntimeError



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from ._json_format_base import _JSONFormatBase


class JSONTransformData(_JSONFormatBase):
    """The translation and rotation of one coordinate system to another."""

    translation: tuple[float, float, float]
    "List of 4 values encoding a quaternion (x, y, z, w)."

    quaternion: tuple[float, float, float, float]
    "List of 3 values encoding the translation vector (x, y, z)"



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from uuid import UUID

import pytest

from raillabel.json_format import JSONObject, JSONFrameInterval, JSONElementDataPointer
from raillabel.format import Object

# == Fixtures =========================


@pytest.fixture
def objects(object_person, object_person_id, object_track, object_track_id) -> dict[UUID, Object]:
    return {
        object_person_id: object_person,
        object_track_id: object_track,
    }


@pytest.fixture
def object_person_json() -> JSONObject:
    return JSONObject(
        name="person_0032",
        type="person",
        frame_intervals=[JSONFrameInterval(frame_start=1, frame_end=1)],
        object_data_pointers={
            "rgb_middle__bbox__person": JSONElementDataPointer(
                frame_intervals=[JSONFrameInterval(frame_start=1, frame_end=1)],
                type="bbox",
                attribute_pointers={
                    "has_red_hat": "boolean",
                    "has_green_hat": "boolean",
                    "number_of_red_clothing_items": "num",
                    "color_of_hat": "text",
                    "clothing_items": "vec",
                },
            ),
            "lidar__cuboid__person": JSONElementDataPointer(
                frame_intervals=[JSONFrameInterval(frame_start=1, frame_end=1)],
                type="cuboid",
                attribute_pointers={
                    "has_red_hat": "boolean",
                    "has_green_hat": "boolean",
                    "number_of_red_clothing_items": "num",
                    "color_of_hat": "text",
                    "clothing_items": "vec",
                },
            ),
            "lidar__vec__person": JSONElementDataPointer(
                frame_intervals=[JSONFrameInterval(frame_start=1, frame_end=1)],
                type="vec",
                attribute_pointers={
                    "has_red_hat": "boolean",
                    "has_green_hat": "boolean",
                    "number_of_red_clothing_items": "num",
                    "color_of_hat": "text",
                    "clothing_items": "vec",
                },
            ),
        },
    )


@pytest.fixture
def object_person() -> Object:
    return Object(
        name="person_0032",
        type="person",
    )


@pytest.fixture
def object_person_id() -> UUID:
    return UUID("b40ba3ad-0327-46ff-9c28-2506cfd6d934")


@pytest.fixture
def object_track_json() -> JSONObject:
    return JSONObject(
        name="track_0001",
        type="track",
        frame_intervals=[JSONFrameInterval(frame_start=1, frame_end=1)],
        object_data_pointers={
            "rgb_middle__poly2d__track": JSONElementDataPointer(
                frame_intervals=[JSONFrameInterval(frame_start=1, frame_end=1)],
                type="poly2d",
                attribute_pointers={
                    "has_red_hat": "boolean",
                    "has_green_hat": "boolean",
                    "number_of_red_clothing_items": "num",
                    "color_of_hat": "text",
                    "clothing_items": "vec",
                },
            ),
            "lidar__poly3d__track": JSONElementDataPointer(
                frame_intervals=[JSONFrameInterval(frame_start=1, frame_end=1)],
                type="poly3d",
                attribute_pointers={
                    "has_red_hat": "boolean",
                    "has_green_hat": "boolean",
                    "number_of_red_clothing_items": "num",
                    "color_of_hat": "text",
                    "clothing_items": "vec",
                },
            ),
        },
    )


@pytest.fixture
def object_track() -> Object:
    return Object(
        name="track_0001",
        type="track",
    )


@pytest.fixture
def object_track_id() -> UUID:
    return UUID("cfcf9750-3BC3-4077-9079-a82c0c63976a")


# == Tests ============================


def test_from_json__person(object_person, object_person_json):
    actual = Object.from_json(object_person_json)
    assert actual == object_person


def test_from_json__track(object_track, object_track_json):
    actual = Object.from_json(object_track_json)
    assert actual == object_track


def test_to_json__person(object_person, object_person_json, object_person_id, frame):
    actual = object_person.to_json(object_person_id, {1: frame})
    assert actual == object_person_json


def test_to_json__track(object_track, object_track_json, object_track_id, frame):
    actual = object_track.to_json(object_track_id, {1: frame})
    assert actual == object_track_json


if __name__ == "__main__":
    pytest.main([__file__, "-vv"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import pytest

from raillabel.format import Quaternion

# == Fixtures =========================


@pytest.fixture
def quaternion_json() -> dict:
    return [0.75318325, -0.10270147, 0.21430262, -0.61338551]


@pytest.fixture
def quaternion() -> dict:
    return Quaternion(0.75318325, -0.10270147, 0.21430262, -0.61338551)


# == Tests ============================


def test_from_json(quaternion, quaternion_json):
    actual = Quaternion.from_json(quaternion_json)
    assert actual == quaternion


def test_to_json(quaternion, quaternion_json):
    actual = quaternion.to_json()
    assert actual == tuple(quaternion_json)


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from dataclasses import dataclass
from decimal import Decimal

from raillabel.format import Frame

from ._filter_abc import _FrameLevelFilter


@dataclass
class StartTimeFilter(_FrameLevelFilter):
    """Filter out all the frames in the scene with timestamps lower than the start_time."""

    start_time: float | Decimal

    def passes_filter(self, _: int, frame: Frame) -> bool:
        """Assess if a frame passes this filter."""
        if frame.timestamp is not None:
            return frame.timestamp >= self.start_time

        return True



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from dataclasses import dataclass

from raillabel.json_format import JSONIntrinsicsRadar


@dataclass
class IntrinsicsRadar:
    """Intrinsic calibration for a radar sensor."""

    resolution_px_per_m: float
    """Factor for calculating the 3D coordinates of a pixel in the cartesian radar images. Number of
    pixels in the images per meter from the sensor."""

    width_px: int
    "Width of the cartesian image frame in pixel."

    height_px: int
    "Height of the cartesian image frame in pixel."

    @classmethod
    def from_json(cls, json: JSONIntrinsicsRadar) -> IntrinsicsRadar:
        """Construct an instant of this class from RailLabel JSON data."""
        return IntrinsicsRadar(
            resolution_px_per_m=json.resolution_px_per_m,
            width_px=json.width_px,
            height_px=json.height_px,
        )

    def to_json(self) -> JSONIntrinsicsRadar:
        """Export this object into the RailLabel JSON format."""
        return JSONIntrinsicsRadar(
            resolution_px_per_m=self.resolution_px_per_m,
            width_px=self.width_px,
            height_px=self.height_px,
        )



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from uuid import UUID

import pytest

from raillabel.format import Num
from raillabel.json_format import JSONNum

# == Fixtures =========================


@pytest.fixture
def num_json(num_id) -> JSONNum:
    return JSONNum(
        uid=num_id,
        name="velocity",
        val=49.21321,
        coordinate_system="gps_imu",
    )


@pytest.fixture
def num_id() -> UUID:
    return UUID("78f0ad89-2750-4a30-9d66-44c9da73a714")


@pytest.fixture
def num(num_id) -> Num:
    return Num(
        sensor_id="gps_imu",
        name="velocity",
        val=49.21321,
        id=num_id,
    )


# == Tests ============================


def test_from_json(num, num_json):
    actual = Num.from_json(num_json)
    assert actual == num


def test_to_json(num, num_json):
    actual = num.to_json()
    assert actual == num_json


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from pathlib import Path

from raillabel.format import Scene


def save(scene: Scene, path: Path | str, prettify_json: bool = False) -> None:
    """Save a raillabel.Scene to a JSON file."""
    if prettify_json:
        json_data = scene.to_json().model_dump_json(exclude_none=True, indent=4)
    else:
        json_data = scene.to_json().model_dump_json(exclude_none=True)

    with Path(path).open("w") as scene_file:
        scene_file.write(json_data)



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0
from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING
from uuid import UUID

if TYPE_CHECKING:
    from raillabel.format import Bbox, Cuboid, Frame, Poly2d, Poly3d, Scene, Seg3d


class _FilterAbc(ABC):
    """Base class of all filter classes regardless of level."""


class _AnnotationLevelFilter(_FilterAbc):
    """Base class of all filter classes applied to the annotations."""

    @abstractmethod
    def passes_filter(
        self, annotation_id: UUID, annotation: Bbox | Cuboid | Poly2d | Poly3d | Seg3d, scene: Scene
    ) -> bool:
        """Assess if an annotation passes this filter."""
        raise NotImplementedError


class _FrameLevelFilter(_FilterAbc):
    """Base class of all filter classes applied to the frames."""

    @abstractmethod
    def passes_filter(self, frame_id: int, frame: Frame) -> bool:
        """Assess if a frame passes this filter."""
        raise NotImplementedError



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING
from uuid import UUID

from raillabel.json_format import JSONElementDataPointer, JSONFrameInterval, JSONObject

from ._attributes import _attributes_to_json
from .frame_interval import FrameInterval

if TYPE_CHECKING:
    from .frame import Frame


@dataclass
class Object:
    """Physical, unique object in the data, that can be tracked via its UID."""

    name: str
    """Name of the object. It is a friendly name and not used for indexing. Commonly the class name
    is used followed by an underscore and an integer (i.e. person_0032)."""

    type: str
    "The type of an object defines the class the object corresponds to (like 'person')."

    @classmethod
    def from_json(cls, json: JSONObject) -> Object:
        """Construct an instant of this class from RailLabel JSON data."""
        return Object(
            name=json.name,
            type=json.type,
        )

    def to_json(self, object_id: UUID, frames: dict[int, Frame]) -> JSONObject:
        """Export this object into the RailLabel JSON format."""
        return JSONObject(
            name=self.name,
            type=self.type,
            frame_intervals=_frame_intervals_to_json(object_id, frames),
            object_data_pointers=_object_data_pointers_to_json(object_id, self.type, frames),
        )


def _frame_intervals_to_json(object_id: UUID, frames: dict[int, Frame]) -> list[JSONFrameInterval]:
    frames_with_this_object = set()

    for frame_id, frame in frames.items():
        for annotation in frame.annotations.values():
            if annotation.object_id == object_id:
                frames_with_this_object.add(frame_id)
                continue

    return [fi.to_json() for fi in FrameInterval.from_frame_ids(list(frames_with_this_object))]


def _object_data_pointers_to_json(
    object_id: UUID, object_type: str, frames: dict[int, Frame]
) -> dict[str, JSONElementDataPointer]:
    pointers_raw = {}

    for frame_id, frame in frames.items():
        for annotation in [ann for ann in frame.annotations.values() if ann.object_id == object_id]:
            annotation_name = annotation.name(object_type)
            if annotation_name not in pointers_raw:
                pointers_raw[annotation_name] = {
                    "frame_intervals": set(),
                    "type": annotation_name.split("__")[1],
                    "attribute_pointers": {},
                }

            pointers_raw[annotation_name]["frame_intervals"].add(frame_id)  # type: ignore
            json_attributes = _attributes_to_json(annotation.attributes)

            if json_attributes is None:
                continue

            for attribute in json_attributes.boolean:  # type: ignore
                pointers_raw[annotation_name]["attribute_pointers"][attribute.name] = "boolean"  # type: ignore

            for attribute in json_attributes.num:  # type: ignore
                pointers_raw[annotation_name]["attribute_pointers"][attribute.name] = "num"  # type: ignore

            for attribute in json_attributes.text:  # type: ignore
                pointers_raw[annotation_name]["attribute_pointers"][attribute.name] = "text"  # type: ignore

            for attribute in json_attributes.vec:  # type: ignore
                pointers_raw[annotation_name]["attribute_pointers"][attribute.name] = "vec"  # type: ignore

    object_data_pointers = {}
    for annotation_name, object_data_pointer in pointers_raw.items():
        object_data_pointers[annotation_name] = JSONElementDataPointer(
            type=object_data_pointer["type"],
            frame_intervals=[
                fi.to_json()
                for fi in FrameInterval.from_frame_ids(list(object_data_pointer["frame_intervals"]))  # type: ignore
            ],
            attribute_pointers=object_data_pointer["attribute_pointers"],
        )

    return object_data_pointers



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from dataclasses import dataclass

from raillabel.json_format import JSONTransformData

from .point3d import Point3d
from .quaternion import Quaternion


@dataclass
class Transform:
    """A transformation between two coordinate systems."""

    pos: Point3d
    "Translation with regards to the parent coordinate system."

    quat: Quaternion
    "Rotation quaternion with regards to the parent coordinate system."

    @classmethod
    def from_json(cls, json: JSONTransformData) -> Transform:
        """Construct an instant of this class from RailLabel JSON data."""
        return Transform(
            pos=Point3d.from_json(json.translation),
            quat=Quaternion.from_json(json.quaternion),
        )

    def to_json(self) -> JSONTransformData:
        """Export this object into the RailLabel JSON format."""
        return JSONTransformData(
            translation=self.pos.to_json(),
            quaternion=self.quat.to_json(),
        )



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from dataclasses import dataclass
from uuid import UUID

from raillabel.format import Bbox, Cuboid, Poly2d, Poly3d, Scene, Seg3d

from ._filter_abc import _AnnotationLevelFilter


@dataclass
class ExcludeAttributesFilter(_AnnotationLevelFilter):
    """Filter out all annotations in the scene, that do have matching attributes.

    If an attribute has None as the value, all annotations are included, that do not have this
    attribute. If the value is anything other than None, all annotations are included that do not
    have the attribute or where the value does not match.
    """

    attributes: dict[str, bool | float | str | list | None]

    def passes_filter(
        self, _: UUID, annotation: Bbox | Cuboid | Poly2d | Poly3d | Seg3d, __: Scene
    ) -> bool:
        """Assess if an annotation passes this filter."""
        for attribute_name, attribute_value in self.attributes.items():
            if attribute_value is None:
                if attribute_name in annotation.attributes:
                    return False

            elif (
                attribute_name in annotation.attributes
                and attribute_value == annotation.attributes[attribute_name]
            ):
                return False

        return True



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import pytest

import raillabel
from raillabel import Scene
from raillabel.json_format import JSONScene


def test_save(json_data, tmp_path):
    scene_path = tmp_path / "scene.json"
    ground_truth_scene = Scene.from_json(JSONScene(**json_data["openlabel_v1_short"]))
    raillabel.save(ground_truth_scene, scene_path)

    actual = raillabel.load(scene_path)
    assert actual == ground_truth_scene


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import pytest

from raillabel.format import Point3d

# == Fixtures =========================


@pytest.fixture
def point3d_json() -> dict:
    return [419, 3.14, 0]


@pytest.fixture
def point3d() -> dict:
    return Point3d(419, 3.14, 0)


@pytest.fixture
def another_point3d_json() -> dict:
    return [419.2, 3.34, 0.2]


@pytest.fixture
def another_point3d() -> dict:
    return Point3d(419.2, 3.34, 0.2)


# == Tests ============================


def test_from_json(point3d, point3d_json):
    actual = Point3d.from_json(point3d_json)
    assert actual == point3d


def test_from_json__another(another_point3d, another_point3d_json):
    actual = Point3d.from_json(another_point3d_json)
    assert actual == another_point3d


def test_to_json(point3d, point3d_json):
    actual = point3d.to_json()
    assert actual == tuple(point3d_json)


def test_to_json__another(another_point3d, another_point3d_json):
    actual = another_point3d.to_json()
    assert actual == tuple(another_point3d_json)


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from dataclasses import dataclass
from uuid import UUID

from raillabel.json_format import JSONNum


@dataclass
class Num:
    """A number."""

    name: str
    "Human readable name describing the annotation."

    val: float
    "This is the value of the number object."

    id: UUID | None = None
    "The unique identifyer of the Num."

    sensor_id: str | None = None
    "A reference to the sensor, this value is represented in."

    @classmethod
    def from_json(cls, json: JSONNum) -> Num:
        """Construct an instant of this class from RailLabel JSON data."""
        return Num(
            name=json.name,
            val=json.val,
            id=json.uid,
            sensor_id=json.coordinate_system,
        )

    def to_json(self) -> JSONNum:
        """Export this object into the RailLabel JSON format."""
        return JSONNum(
            name=self.name,
            val=self.val,
            coordinate_system=self.sensor_id,
            uid=self.id,
        )



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from dataclasses import dataclass
from uuid import UUID

from raillabel.format import Bbox, Cuboid, Poly2d, Poly3d, Scene, Seg3d

from ._filter_abc import _AnnotationLevelFilter


@dataclass
class IncludeSensorIdFilter(_AnnotationLevelFilter):
    """Filter out all annotations in the scene, that do NOT have matching sensor ids."""

    sensor_ids: set[str] | list[str]

    def passes_filter(
        self, _: UUID, annotation: Bbox | Cuboid | Poly2d | Poly3d | Seg3d, __: Scene
    ) -> bool:
        """Assess if an annotation passes this filter."""
        return annotation.sensor_id in self.sensor_ids



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from uuid import UUID

import pytest

from raillabel.format import Poly2d
from raillabel.json_format import JSONPoly2d

# == Fixtures =========================


@pytest.fixture
def poly2d_json(
    point2d_json,
    another_point2d_json,
    attributes_multiple_types_json,
) -> JSONPoly2d:
    return JSONPoly2d(
        uid="013e7b34-62E5-435c-9412-87318c50f6d8",
        name="rgb_middle__poly2d__track",
        closed=True,
        mode="MODE_POLY2D_ABSOLUTE",
        val=point2d_json + another_point2d_json,
        coordinate_system="rgb_middle",
        attributes=attributes_multiple_types_json,
    )


@pytest.fixture
def poly2d_id() -> UUID:
    return UUID("013e7b34-62E5-435c-9412-87318c50f6d8")


@pytest.fixture
def poly2d(
    point2d,
    another_point2d,
    attributes_multiple_types,
    object_track_id,
) -> Poly2d:
    return Poly2d(
        points=[point2d, another_point2d],
        closed=True,
        sensor_id="rgb_middle",
        attributes=attributes_multiple_types,
        object_id=object_track_id,
    )


# == Tests ============================


def test_from_json(poly2d, poly2d_json, object_track_id):
    actual = Poly2d.from_json(poly2d_json, object_track_id)
    assert actual == poly2d


def test_name(poly2d):
    actual = poly2d.name("track")
    assert actual == "rgb_middle__poly2d__track"


def test_to_json(poly2d, poly2d_json, poly2d_id):
    actual = poly2d.to_json(poly2d_id, object_type="track")
    assert actual == poly2d_json


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import pytest

from raillabel.format import OtherSensor
from raillabel.json_format import JSONCoordinateSystem, JSONStreamOther

# == Fixtures =========================


@pytest.fixture
def other_json(transform_json) -> tuple[JSONStreamOther, JSONCoordinateSystem]:
    return (
        JSONStreamOther(
            type="other",
            uri="/path/to/sensor/data",
            description="A very nice generic sensor",
        ),
        JSONCoordinateSystem(
            parent="base", type="sensor", pose_wrt_parent=transform_json, children=None
        ),
    )


@pytest.fixture
def other(transform) -> dict:
    return OtherSensor(
        extrinsics=transform,
        uri="/path/to/sensor/data",
        description="A very nice generic sensor",
    )


# == Tests ============================


def test_from_json(other, other_json):
    actual = OtherSensor.from_json(other_json[0], other_json[1])
    assert actual == other


def test_to_json(other, other_json):
    actual = other.to_json()
    assert actual == other_json


if __name__ == "__main__":
    pytest.main([__file__, "--disable-pytest-warnings", "--cache-clear", "-v"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from dataclasses import dataclass


@dataclass
class Size2d:
    """The size of a rectangle in a 2d image."""

    x: float
    "The size along the x-axis."

    y: float
    "The size along the y-axis."

    @classmethod
    def from_json(cls, json: tuple[float, float]) -> Size2d:
        """Construct an instant of this class from RailLabel JSON data."""
        return Size2d(x=json[0], y=json[1])

    def to_json(self) -> tuple[float, float]:
        """Export this object into the RailLabel JSON format."""
        return (self.x, self.y)



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from dataclasses import dataclass
from uuid import UUID

from raillabel.json_format import JSONCuboid

from ._attributes import _attributes_from_json, _attributes_to_json
from .point3d import Point3d
from .quaternion import Quaternion
from .size3d import Size3d


@dataclass
class Cuboid:
    """3D bounding box."""

    pos: Point3d
    """The center position of the cuboid in meters, where the x coordinate points ahead of the
    vehicle, y points to the left and z points upwards."""

    quat: Quaternion
    "The rotation of the cuboid in quaternions."

    size: Size3d
    "The size of the cuboid in meters."

    object_id: UUID
    "The unique identifyer of the real-life object, this annotation belongs to."

    sensor_id: str
    "The unique identifyer of the sensor this annotation is labeled in."

    attributes: dict[str, float | bool | str | list]
    "Additional information associated with the annotation."

    @classmethod
    def from_json(cls, json: JSONCuboid, object_id: UUID) -> Cuboid:
        """Construct an instant of this class from RailLabel JSON data."""
        return Cuboid(
            pos=Point3d.from_json((json.val[0], json.val[1], json.val[2])),
            quat=Quaternion.from_json((json.val[3], json.val[4], json.val[5], json.val[6])),
            size=Size3d.from_json((json.val[7], json.val[8], json.val[9])),
            object_id=object_id,
            sensor_id=json.coordinate_system,
            attributes=_attributes_from_json(json.attributes),
        )

    def to_json(self, uid: UUID, object_type: str) -> JSONCuboid:
        """Export this object into the RailLabel JSON format."""
        return JSONCuboid(
            name=self.name(object_type),
            val=list(self.pos.to_json()) + list(self.quat.to_json()) + list(self.size.to_json()),
            coordinate_system=self.sensor_id,
            uid=uid,
            attributes=_attributes_to_json(self.attributes),
        )

    def name(self, object_type: str) -> str:
        """Return the name of the annotation used for indexing in the object data pointers."""
        return f"{self.sensor_id}__cuboid__{object_type}"



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import pytest

from raillabel.format import IntrinsicsRadar
from raillabel.json_format import JSONIntrinsicsRadar

# == Fixtures =========================


@pytest.fixture
def intrinsics_radar_json() -> dict:
    return JSONIntrinsicsRadar(
        resolution_px_per_m=2.856,
        width_px=2856,
        height_px=1428,
    )


@pytest.fixture
def intrinsics_radar() -> dict:
    return IntrinsicsRadar(
        resolution_px_per_m=2.856,
        width_px=2856,
        height_px=1428,
    )


# == Tests ============================


def test_from_json(intrinsics_radar, intrinsics_radar_json):
    actual = IntrinsicsRadar.from_json(intrinsics_radar_json)
    assert actual == intrinsics_radar


def test_to_json(intrinsics_radar, intrinsics_radar_json):
    actual = intrinsics_radar.to_json()
    assert actual == intrinsics_radar_json


if __name__ == "__main__":
    pytest.main([__file__, "--disable-pytest-warnings", "--cache-clear", "-v"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from uuid import UUID

import pytest

import raillabel
from raillabel.scene_builder import SceneBuilder


def test_include_frame_ids():
    scene = SceneBuilder.empty().add_frame(1).add_frame(2).add_frame(3).result
    filters = [raillabel.filter.IncludeFrameIdFilter([1, 3])]

    actual = scene.filter(filters)
    assert actual == SceneBuilder.empty().add_frame(1).add_frame(3).result


def test_exclude_frame_ids():
    scene = SceneBuilder.empty().add_frame(1).add_frame(2).add_frame(3).result
    filters = [raillabel.filter.ExcludeFrameIdFilter([2])]

    actual = scene.filter(filters)
    assert actual == SceneBuilder.empty().add_frame(1).add_frame(3).result


def test_start_time():
    scene = SceneBuilder.empty().add_frame(1, 100).add_frame(2, 200).add_frame(3, 300).result
    filters = [raillabel.filter.StartTimeFilter(150)]

    actual = scene.filter(filters)
    assert actual == SceneBuilder.empty().add_frame(2, 200).add_frame(3, 300).result


def test_end_time():
    scene = SceneBuilder.empty().add_frame(1, 100).add_frame(2, 200).add_frame(3, 300).result
    filters = [raillabel.filter.EndTimeFilter(250)]

    actual = scene.filter(filters)
    assert actual == SceneBuilder.empty().add_frame(1, 100).add_frame(2, 200).result


def test_include_annotation_ids():
    scene = (
        SceneBuilder.empty()
        .add_bbox(uid="6c95543d-0000-4000-0000-000000000000")
        .add_bbox(uid="6c95543d-0000-4000-0000-000000000001")
        .add_bbox(uid="6c95543d-0000-4000-0000-000000000002")
        .result
    )
    filters = [
        raillabel.filter.IncludeAnnotationIdFilter(
            [
                UUID("6c95543d-0000-4000-0000-000000000000"),
                UUID("6c95543d-0000-4000-0000-000000000002"),
            ]
        )
    ]

    actual = scene.filter(filters)
    assert (
        actual
        == SceneBuilder.empty()
        .add_bbox(uid="6c95543d-0000-4000-0000-000000000000")
        .add_bbox(uid="6c95543d-0000-4000-0000-000000000002")
        .result
    )


def test_exclude_annotation_ids():
    scene = (
        SceneBuilder.empty()
        .add_bbox(uid="6c95543d-0000-4000-0000-000000000000")
        .add_bbox(uid="6c95543d-0000-4000-0000-000000000001")
        .add_bbox(uid="6c95543d-0000-4000-0000-000000000002")
        .result
    )
    filters = [
        raillabel.filter.ExcludeAnnotationIdFilter([UUID("6c95543d-0000-4000-0000-000000000001")])
    ]

    actual = scene.filter(filters)
    assert (
        actual
        == SceneBuilder.empty()
        .add_bbox(uid="6c95543d-0000-4000-0000-000000000000")
        .add_bbox(uid="6c95543d-0000-4000-0000-000000000002")
        .result
    )


def test_include_annotation_type():
    scene = SceneBuilder.empty().add_bbox().add_cuboid().result
    filters = [raillabel.filter.IncludeAnnotationTypeFilter(["bbox"])]

    actual = scene.filter(filters)
    assert actual == SceneBuilder.empty().add_bbox().result


def test_exclude_annotation_type():
    scene = SceneBuilder.empty().add_bbox().add_cuboid().result
    filters = [raillabel.filter.ExcludeAnnotationTypeFilter(["cuboid"])]

    actual = scene.filter(filters)
    assert actual == SceneBuilder.empty().add_bbox().result


def test_include_object_ids():
    scene = (
        SceneBuilder.empty()
        .add_bbox(object_name="person_0001")
        .add_cuboid(object_name="train_0001")
        .result
    )
    filters = [
        raillabel.filter.IncludeObjectIdFilter([UUID("5c59aad4-0000-4000-0000-000000000000")])
    ]

    actual = scene.filter(filters)
    assert actual == SceneBuilder.empty().add_bbox(object_name="person_0001").result


def test_exclude_object_ids():
    scene = (
        SceneBuilder.empty()
        .add_bbox(object_name="person_0001")
        .add_cuboid(object_name="train_0001")
        .result
    )
    filters = [
        raillabel.filter.ExcludeObjectIdFilter([UUID("5c59aad4-0000-4000-0000-000000000001")])
    ]

    actual = scene.filter(filters)
    assert actual == SceneBuilder.empty().add_bbox(object_name="person_0001").result


def test_include_object_types():
    scene = (
        SceneBuilder.empty()
        .add_bbox(object_name="person_0001")
        .add_cuboid(object_name="train_0001")
        .result
    )
    filters = [raillabel.filter.IncludeObjectTypeFilter(["person"])]

    actual = scene.filter(filters)
    assert actual == SceneBuilder.empty().add_bbox(object_name="person_0001").result


def test_exclude_object_types():
    scene = (
        SceneBuilder.empty()
        .add_bbox(object_name="person_0001")
        .add_cuboid(object_name="train_0001")
        .result
    )
    filters = [raillabel.filter.ExcludeObjectTypeFilter(["train"])]

    actual = scene.filter(filters)
    assert actual == SceneBuilder.empty().add_bbox(object_name="person_0001").result


def test_include_sensor_ids():
    scene = (
        SceneBuilder.empty().add_bbox(sensor_id="rgb_middle").add_cuboid(sensor_id="lidar").result
    )
    filters = [raillabel.filter.IncludeSensorIdFilter(["rgb_middle"])]

    actual = scene.filter(filters)
    assert actual == SceneBuilder.empty().add_bbox(sensor_id="rgb_middle").result


def test_exclude_sensor_ids():
    scene = (
        SceneBuilder.empty().add_bbox(sensor_id="rgb_middle").add_cuboid(sensor_id="lidar").result
    )
    filters = [raillabel.filter.ExcludeSensorIdFilter(["lidar"])]

    actual = scene.filter(filters)
    assert actual == SceneBuilder.empty().add_bbox(sensor_id="rgb_middle").result


def test_include_sensor_types():
    scene = (
        SceneBuilder.empty()
        .add_bbox(sensor_id="rgb_middle")
        .add_cuboid(sensor_id="lidar_middle")
        .result
    )
    filters = [raillabel.filter.IncludeSensorTypeFilter(["camera"])]

    actual = scene.filter(filters)
    assert actual == SceneBuilder.empty().add_bbox(sensor_id="rgb_middle").result


def test_exclude_sensor_types():
    scene = (
        SceneBuilder.empty()
        .add_bbox(sensor_id="rgb_middle")
        .add_cuboid(sensor_id="lidar_middle")
        .result
    )
    filters = [raillabel.filter.ExcludeSensorTypeFilter(["lidar"])]

    actual = scene.filter(filters)
    assert actual == SceneBuilder.empty().add_bbox(sensor_id="rgb_middle").result


def test_include_attributes__only_keys():
    scene = (
        SceneBuilder.empty()
        .add_bbox(attributes={"length": 42})
        .add_bbox(attributes={"width": 34})
        .result
    )
    filters = [raillabel.filter.IncludeAttributesFilter({"length": None})]

    actual = scene.filter(filters)
    assert actual == SceneBuilder.empty().add_bbox(attributes={"length": 42}).result


def test_include_attributes__with_values():
    scene = (
        SceneBuilder.empty()
        .add_bbox(attributes={"is_dummy": True})
        .add_bbox(attributes={"is_dummy": False})
        .result
    )
    filters = [raillabel.filter.IncludeAttributesFilter({"is_dummy": True})]

    actual = scene.filter(filters)
    assert actual == SceneBuilder.empty().add_bbox(attributes={"is_dummy": True}).result


def test_exclude_attributes__only_keys():
    scene = (
        SceneBuilder.empty()
        .add_bbox(attributes={"length": 42})
        .add_bbox(attributes={"width": 34})
        .result
    )
    filters = [raillabel.filter.ExcludeAttributesFilter({"width": None})]

    actual = scene.filter(filters)
    assert actual == SceneBuilder.empty().add_bbox(attributes={"length": 42}).result


def test_exclude_attributes__with_values():
    scene = (
        SceneBuilder.empty()
        .add_bbox(attributes={"is_dummy": True})
        .add_bbox(attributes={"is_dummy": False})
        .result
    )
    filters = [raillabel.filter.ExcludeAttributesFilter({"is_dummy": False})]

    actual = scene.filter(filters)
    assert actual == SceneBuilder.empty().add_bbox(attributes={"is_dummy": True}).result


if __name__ == "__main__":
    pytest.main([__file__, "-vv"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from ._json_format_base import _JSONFormatBase


class JSONFrameInterval(_JSONFormatBase):
    """A frame interval defines a starting and ending frame number as a closed interval.

    That means the interval includes the limit frame numbers.
    """

    frame_start: int
    "Initial frame number of the interval."

    frame_end: int
    "Ending frame number of the interval."



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import pytest
from decimal import Decimal

from raillabel.format import SensorReference
from raillabel.json_format import JSONStreamSync, JSONStreamSyncProperties, JSONStreamSyncTimestamp

# == Fixtures =========================


@pytest.fixture
def sensor_reference_json() -> JSONStreamSync:
    return JSONStreamSync(
        stream_properties=JSONStreamSyncProperties(
            sync=JSONStreamSyncTimestamp(timestamp="1631337747.123123123")
        ),
        uri="/uri/to/file.png",
    )


@pytest.fixture
def sensor_reference() -> SensorReference:
    return SensorReference(timestamp=Decimal("1631337747.123123123"), uri="/uri/to/file.png")


@pytest.fixture
def another_sensor_reference_json() -> JSONStreamSync:
    return JSONStreamSync(
        stream_properties=JSONStreamSyncProperties(
            sync=JSONStreamSyncTimestamp(timestamp="1631337747.103123123")
        ),
        uri="/uri/to/file.pcd",
    )


@pytest.fixture
def another_sensor_reference() -> SensorReference:
    return SensorReference(timestamp=Decimal("1631337747.103123123"), uri="/uri/to/file.pcd")


# == Tests ============================


def test_from_json(sensor_reference, sensor_reference_json):
    actual = SensorReference.from_json(sensor_reference_json)
    assert actual == sensor_reference


def test_from_json__another(another_sensor_reference, another_sensor_reference_json):
    actual = SensorReference.from_json(another_sensor_reference_json)
    assert actual == another_sensor_reference


def test_to_json(sensor_reference, sensor_reference_json):
    actual = sensor_reference.to_json()
    assert actual == sensor_reference_json


def test_to_json__another(another_sensor_reference, another_sensor_reference_json):
    actual = another_sensor_reference.to_json()
    assert actual == another_sensor_reference_json


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import pytest

from raillabel.format import Camera, IntrinsicsPinhole
from raillabel.json_format import (
    JSONCoordinateSystem,
    JSONStreamCamera,
    JSONStreamCameraProperties,
    JSONIntrinsicsPinhole,
)

# == Fixtures =========================


@pytest.fixture
def camera_json(
    intrinsics_pinhole_json, transform_json
) -> tuple[JSONStreamCamera, JSONCoordinateSystem]:
    return (
        JSONStreamCamera(
            type="camera",
            stream_properties=JSONStreamCameraProperties(intrinsics_pinhole=intrinsics_pinhole_json),
            uri="/path/to/sensor/data",
            description="A very nice camera",
        ),
        JSONCoordinateSystem(
            parent="base", type="sensor", pose_wrt_parent=transform_json, children=None
        ),
    )


@pytest.fixture
def camera(intrinsics_pinhole, transform) -> Camera:
    return Camera(
        intrinsics=intrinsics_pinhole,
        extrinsics=transform,
        uri="/path/to/sensor/data",
        description="A very nice camera",
    )


@pytest.fixture
def camera_empty() -> Camera:
    return Camera(
        intrinsics=IntrinsicsPinhole(
            camera_matrix=tuple([0] * 12),
            distortion=tuple([0] * 5),
            width_px=0,
            height_px=0,
        )
    )


# == Tests ============================


def test_from_json(camera, camera_json):
    actual = Camera.from_json(camera_json[0], camera_json[1])
    assert actual == camera


def test_to_json(camera, camera_json):
    actual = camera.to_json()
    assert actual == camera_json


if __name__ == "__main__":
    pytest.main([__file__, "-vv"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from typing import Literal

from ._json_format_base import _JSONFormatBase


class JSONStreamRadar(_JSONFormatBase):
    """A stream describes the source of a data sequence, usually a sensor.

    This specific object contains the intrinsics of a radar sensor.
    """

    type: Literal["radar"]
    "A string encoding the type of the stream."

    stream_properties: JSONStreamRadarProperties
    "Intrinsic calibration of the stream."

    uri: str
    "A string encoding the subdirectory containing the sensor files."

    description: str | None = None
    "Description of the stream."


class JSONStreamRadarProperties(_JSONFormatBase):
    """Intrinsic calibration of the stream."""

    intrinsics_radar: JSONIntrinsicsRadar


class JSONIntrinsicsRadar(_JSONFormatBase):
    """JSON object defining an instance of the intrinsic parameters of a radar."""

    resolution_px_per_m: float
    "Number correlating pixel in the output image to a position relative to the sensor in meters."

    height_px: int
    "Height of the radar output in pixel."

    width_px: int
    "Width of the radar output in pixel."



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from ._json_format_base import _JSONFormatBase
from .boolean_attribute import JSONBooleanAttribute
from .num_attribute import JSONNumAttribute
from .text_attribute import JSONTextAttribute
from .vec_attribute import JSONVecAttribute


class JSONAttributes(_JSONFormatBase):
    """Attributes is the alias of element data that can be nested inside geometric object data.

    For example, a certain bounding box can have attributes related to its score, visibility, etc.
    These values can be nested inside the bounding box as attributes.
    """

    boolean: list[JSONBooleanAttribute] | None = None
    num: list[JSONNumAttribute] | None = None
    text: list[JSONTextAttribute] | None = None
    vec: list[JSONVecAttribute] | None = None



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0
"""Module containing all relevant format classes."""

from .bbox import Bbox
from .camera import Camera
from .cuboid import Cuboid
from .frame import Frame
from .frame_interval import FrameInterval
from .gps_imu import GpsImu
from .intrinsics_pinhole import IntrinsicsPinhole
from .intrinsics_radar import IntrinsicsRadar
from .lidar import Lidar
from .metadata import Metadata
from .num import Num
from .object import Object
from .other_sensor import OtherSensor
from .point2d import Point2d
from .point3d import Point3d
from .poly2d import Poly2d
from .poly3d import Poly3d
from .quaternion import Quaternion
from .radar import Radar
from .scene import Scene
from .seg3d import Seg3d
from .sensor_reference import SensorReference
from .size2d import Size2d
from .size3d import Size3d
from .transform import Transform

__all__ = [
    "Bbox",
    "Camera",
    "Cuboid",
    "ElementDataPointer",
    "Frame",
    "FrameInterval",
    "GpsImu",
    "IntrinsicsPinhole",
    "IntrinsicsRadar",
    "Lidar",
    "Metadata",
    "Num",
    "Object",
    "OtherSensor",
    "Point2d",
    "Point3d",
    "Poly2d",
    "Poly3d",
    "Quaternion",
    "Radar",
    "Scene",
    "Seg3d",
    "SensorReference",
    "Size2d",
    "Size3d",
    "Transform",
]



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from decimal import Decimal

from ._json_format_base import _JSONFormatBase


class JSONStreamSync(_JSONFormatBase):
    """Syncronization information of a stream in a frame."""

    stream_properties: JSONStreamSyncProperties
    uri: str | None = None


class JSONStreamSyncProperties(_JSONFormatBase):
    """The sync information."""

    sync: JSONStreamSyncTimestamp


class JSONStreamSyncTimestamp(_JSONFormatBase):
    """The timestamp of a stream sync."""

    timestamp: Decimal | str



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from decimal import Decimal

import pytest

from raillabel.format import Frame
from raillabel.json_format import (
    JSONFrame,
    JSONFrameData,
    JSONFrameProperties,
    JSONObjectData,
    JSONAnnotations,
)

# == Fixtures =========================


@pytest.fixture
def frame_json(
    sensor_reference_json,
    another_sensor_reference_json,
    num_json,
    bbox_json,
    cuboid_json,
    poly2d_json,
    poly3d_json,
    seg3d_json,
) -> JSONFrame:
    return JSONFrame(
        frame_properties=JSONFrameProperties(
            timestamp=Decimal("1631337747.123123123"),
            streams={
                "rgb_middle": sensor_reference_json,
                "lidar": another_sensor_reference_json,
            },
            frame_data=JSONFrameData(num=[num_json]),
        ),
        objects={
            "cfcf9750-3bc3-4077-9079-a82c0c63976a": JSONObjectData(
                object_data=JSONAnnotations(
                    poly2d=[poly2d_json],
                    poly3d=[poly3d_json],
                )
            ),
            "b40ba3ad-0327-46ff-9c28-2506cfd6d934": JSONObjectData(
                object_data=JSONAnnotations(
                    bbox=[bbox_json],
                    cuboid=[cuboid_json],
                    vec=[seg3d_json],
                )
            ),
        },
    )


@pytest.fixture
def frame(
    sensor_reference,
    another_sensor_reference,
    num,
    bbox,
    bbox_id,
    cuboid,
    cuboid_id,
    poly2d,
    poly2d_id,
    poly3d,
    poly3d_id,
    seg3d,
    seg3d_id,
) -> dict:
    return Frame(
        timestamp=Decimal("1631337747.123123123"),
        sensors={
            "rgb_middle": sensor_reference,
            "lidar": another_sensor_reference,
        },
        frame_data={num.name: num},
        annotations={
            bbox_id: bbox,
            cuboid_id: cuboid,
            poly2d_id: poly2d,
            poly3d_id: poly3d,
            seg3d_id: seg3d,
        },
    )


# == Tests ============================


def test_from_json(frame, frame_json):
    actual = Frame.from_json(frame_json)
    assert actual == frame


def test_to_json(frame, frame_json, objects):
    actual = frame.to_json(objects)
    assert actual == frame_json


if __name__ == "__main__":
    pytest.main([__file__, "-vv"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import pytest

from raillabel.format import Scene
from raillabel.json_format import (
    JSONScene,
    JSONSceneContent,
    JSONCoordinateSystem,
    JSONFrameInterval,
)

# == Fixtures =========================


@pytest.fixture
def scene_json(
    metadata_json,
    camera_json,
    lidar_json,
    radar_json,
    object_person_id,
    object_person_json,
    object_track_id,
    object_track_json,
    frame_json,
) -> JSONScene:
    return JSONScene(
        openlabel=JSONSceneContent(
            metadata=metadata_json,
            coordinate_systems={
                "base": JSONCoordinateSystem(
                    parent="",
                    type="local",
                    pose_wrt_parent=None,
                    children=["rgb_middle", "lidar", "radar"],
                ),
                "rgb_middle": camera_json[1],
                "lidar": lidar_json[1],
                "radar": radar_json[1],
            },
            streams={
                "rgb_middle": camera_json[0],
                "lidar": lidar_json[0],
                "radar": radar_json[0],
            },
            objects={
                object_person_id: object_person_json,
                object_track_id: object_track_json,
            },
            frames={1: frame_json},
            frame_intervals=[JSONFrameInterval(frame_start=1, frame_end=1)],
        )
    )


@pytest.fixture
def scene(
    metadata,
    camera,
    lidar,
    radar,
    object_person_id,
    object_person,
    object_track_id,
    object_track,
    frame,
) -> Scene:
    return Scene(
        metadata=metadata,
        sensors={
            "rgb_middle": camera,
            "lidar": lidar,
            "radar": radar,
        },
        objects={
            object_person_id: object_person,
            object_track_id: object_track,
        },
        frames={1: frame},
    )


# == Tests ============================


def test_from_json(scene, scene_json):
    actual = Scene.from_json(scene_json)
    assert actual == scene


def test_to_json(scene, scene_json):
    actual = scene.to_json()
    assert actual == scene_json


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from ._json_format_base import _JSONFormatBase
from .bbox import JSONBbox
from .cuboid import JSONCuboid
from .poly2d import JSONPoly2d
from .poly3d import JSONPoly3d
from .vec import JSONVec


class JSONObjectData(_JSONFormatBase):
    """Container of annotations of an object in a frame."""

    object_data: JSONAnnotations


class JSONAnnotations(_JSONFormatBase):
    """Container of the annotations by type."""

    bbox: list[JSONBbox] | None = None
    cuboid: list[JSONCuboid] | None = None
    poly2d: list[JSONPoly2d] | None = None
    poly3d: list[JSONPoly3d] | None = None
    vec: list[JSONVec] | None = None



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from dataclasses import dataclass
from uuid import UUID

from raillabel.json_format import JSONVec

from ._attributes import _attributes_from_json, _attributes_to_json


@dataclass
class Seg3d:
    """The 3D segmentation of a lidar pointcloud."""

    point_ids: list[int]
    "The list of point indices."

    object_id: UUID
    "The unique identifyer of the real-life object, this annotation belongs to."

    sensor_id: str
    "The unique identifyer of the sensor this annotation is labeled in."

    attributes: dict[str, float | bool | str | list]
    "Additional information associated with the annotation."

    @classmethod
    def from_json(cls, json: JSONVec, object_id: UUID) -> Seg3d:
        """Construct an instant of this class from RailLabel JSON data."""
        return Seg3d(
            point_ids=[int(point_id) for point_id in json.val],
            object_id=object_id,
            sensor_id=json.coordinate_system,
            attributes=_attributes_from_json(json.attributes),
        )

    def to_json(self, uid: UUID, object_type: str) -> JSONVec:
        """Export this object into the RailLabel JSON format."""
        return JSONVec(
            name=self.name(object_type),
            val=self.point_ids,
            coordinate_system=self.sensor_id,
            uid=uid,
            attributes=_attributes_to_json(self.attributes),
        )

    def name(self, object_type: str) -> str:
        """Return the name of the annotation used for indexing in the object data pointers."""
        return f"{self.sensor_id}__vec__{object_type}"



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from typing import Literal
from uuid import UUID

from ._json_format_base import _JSONFormatBase
from .attributes import JSONAttributes


class JSONVec(_JSONFormatBase):
    """A vector (list) of numbers."""

    name: str
    """This is a string encoding the name of this object data. It is used as index inside the
    corresponding object data pointers."""

    val: list[float]
    "The numerical values of the vector (list) of numbers."

    coordinate_system: str
    "Name of the coordinate system in respect of which this object data is expressed."

    uid: UUID | None = None
    "This is a string encoding the Universal Unique identifyer of the annotation."

    type: Literal["values", "range"] | None = None
    """This attribute specifies whether the vector shall be considered as a descriptor of individual
    values or as a definition of a range."""

    attributes: JSONAttributes | None = None



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

import pytest

from raillabel.format import Point2d

# == Fixtures =========================


@pytest.fixture
def point2d_json() -> tuple[float, float]:
    return [1.5, 222]


@pytest.fixture
def point2d() -> Point2d:
    return Point2d(1.5, 222)


@pytest.fixture
def another_point2d_json() -> tuple[float, float]:
    return [1.7, 222.2]


@pytest.fixture
def another_point2d() -> Point2d:
    return Point2d(1.7, 222.2)


# == Tests ============================


def test_from_json(point2d, point2d_json):
    actual = Point2d.from_json(point2d_json)
    assert actual == point2d


def test_from_json__another(another_point2d, another_point2d_json):
    actual = Point2d.from_json(another_point2d_json)
    assert actual == another_point2d


def test_to_json(point2d, point2d_json):
    actual = point2d.to_json()
    assert actual == tuple(point2d_json)


def test_to_json__another(another_point2d, another_point2d_json):
    actual = another_point2d.to_json()
    assert actual == tuple(another_point2d_json)


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from uuid import UUID

from ._json_format_base import _JSONFormatBase
from .attributes import JSONAttributes


class JSONPoly3d(_JSONFormatBase):
    """A 3D polyline defined as a sequence of 3D points."""

    name: str
    """This is a string encoding the name of this object data. It is used as index inside the
    corresponding object data pointers."""

    val: list[float]
    "List of numerical values of the polyline, according to its mode."

    closed: bool
    """A boolean that defines whether the polyline is closed or not. In case it is closed, it is
    assumed that the last point of the sequence is connected with the first one."""

    coordinate_system: str
    "Name of the coordinate system in respect of which this object data is expressed."

    uid: UUID | None = None
    "This is a string encoding the Universal Unique identifyer of the annotation."

    attributes: JSONAttributes | None = None



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from raillabel.json_format import (
    JSONAttributes,
    JSONBooleanAttribute,
    JSONNumAttribute,
    JSONTextAttribute,
    JSONVecAttribute,
)


def _attributes_from_json(json: JSONAttributes | None) -> dict[str, float | bool | str | list]:
    """Parse the annotation attributes from json."""
    if json is None:
        return {}

    attributes: dict[str, float | bool | str | list] = {}

    if json.boolean is not None:
        for bool_attribute in json.boolean:
            attributes[bool_attribute.name] = bool_attribute.val

    if json.num is not None:
        for num_attribute in json.num:
            attributes[num_attribute.name] = num_attribute.val

    if json.text is not None:
        for text_attribute in json.text:
            attributes[text_attribute.name] = text_attribute.val

    if json.vec is not None:
        for vec_attribute in json.vec:
            attributes[vec_attribute.name] = vec_attribute.val

    return attributes


def _attributes_to_json(attributes: dict[str, float | bool | str | list]) -> JSONAttributes | None:
    if len(attributes) == 0:
        return None

    boolean_attributes = []
    num_attributes = []
    text_attributes = []
    vec_attributes = []

    for name, value in attributes.items():
        if isinstance(value, bool):
            boolean_attributes.append(JSONBooleanAttribute(name=name, val=value))

        elif isinstance(value, (int, float)):
            num_attributes.append(JSONNumAttribute(name=name, val=value))

        elif isinstance(value, str):
            text_attributes.append(JSONTextAttribute(name=name, val=value))

        elif isinstance(value, list):
            vec_attributes.append(JSONVecAttribute(name=name, val=value))

        else:
            raise UnsupportedAttributeTypeError(name, value)

    return JSONAttributes(
        boolean=boolean_attributes, num=num_attributes, text=text_attributes, vec=vec_attributes
    )


class UnsupportedAttributeTypeError(TypeError):
    def __init__(self, attribute_name: str, attribute_value: object) -> None:
        super().__init__(
            f"{attribute_value.__class__.__name__} of attribute {attribute_name} "
            "is not a supported attribute type"
        )



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from ._json_format_base import _JSONFormatBase


class JSONVecAttribute(_JSONFormatBase):
    """A vec attribute."""

    name: str
    """Friendly identifier describing the attribute. Used to track the attribute throughout
    annotations and frames."""

    val: list[float | str]
    "The value vector of the attribute."



# Copyright DB InfraGO AG and contributors
# SPDX-License-Identifier: Apache-2.0

from __future__ import annotations

from uuid import UUID

from ._json_format_base import _JSONFormatBase
from .coordinate_system import JSONCoordinateSystem
from .frame import JSONFrame
from .frame_interval import JSONFrameInterval
from .metadata import JSONMetadata
from .object import JSONObject
from .stream_camera import JSONStreamCamera
from .stream_other import JSONStreamOther
from .stream_radar import JSONStreamRadar


class JSONScene(_JSONFormatBase):
    """Root RailLabel object."""

    openlabel: JSONSceneContent


class JSONSceneContent(_JSONFormatBase):
    """Container of all scene content."""

    metadata: JSONMetadata
    coordinate_systems: dict[str, JSONCoordinateSystem] | None = None
    streams: dict[str, JSONStreamCamera | JSONStreamOther | JSONStreamRadar] | None = None
    objects: dict[UUID, JSONObject] | None = None
    frames: dict[int, JSONFrame] | None = None
    frame_intervals: list[JSONFrameInterval] | None = None
